#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "ConvFunc.h"
#include "StackFunc.h"
#include "StrFunc.h"
#include "Resources\Resource.h"

#include "AkelEdit.h"
#include "AkelDLL.h"


//Include conversion functions
#define hex2decA
#define hex2decW
#define xatoiA
#define xatoiW
#include "ConvFunc.h"

//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#include "StackFunc.h"

//Include string functions
#define xstrcmpnW
#include "StrFunc.h"

//Defines
#define STRID_GLOBAL_FONTSTYLE_GROUP 1
#define STRID_ENABLE                 2
#define STRID_IGNORE                 3
#define STRID_NORMAL                 4
#define STRID_BOLD                   5
#define STRID_ITALIC                 6
#define STRID_BOLDITALIC             7
#define STRID_AUTOMARK_GROUP         8
#define STRID_SYMBOLS                9
#define STRID_WORDS                  10
#define STRID_TEXTCOLOR              11
#define STRID_BGCOLOR                12
#define STRID_CACHE_ENABLE           13
#define STRID_OK                     14
#define STRID_CANCEL                 15

#define OF_SETTINGS  0x1
#define OF_CACHE     0x2

#define DLLA_HIGHLIGHT_SETEXTENSION     1
#define DLLA_HIGHLIGHT_MARKSELECTION    2
#define DLLA_HIGHLIGHT_UNMARKSELECTION  3
#define DLLA_HIGHLIGHT_CLEARCACHE       4

#define BUFFER_SIZE      1024

#define FF_FONTSTYLE 0x1
#define FF_FONTSIZE  0x2
#define FF_FONTFACE  0x4

#define MARKID_TEXT    1
#define MARKID_AUTO    (DWORD)-1

#define MARKAUTO_NONE    0
#define MARKAUTO_SYMBOLS 1
#define MARKAUTO_WORDS   2

#define MARKMAX_WORD   512

typedef struct _EXTENSIONINFOA {
  struct _EXTENSIONINFOA *next;
  struct _EXTENSIONINFOA *prev;
  char *pExtension;
  int nExtensionLen;
} EXTENSIONINFOA;

typedef struct _EXTENSIONINFOW {
  struct _EXTENSIONINFOW *next;
  struct _EXTENSIONINFOW *prev;
  wchar_t *wpExtension;
  int nExtensionLen;
} EXTENSIONINFOW;

typedef struct _DELIMITERINFOA {
  struct _DELIMITERINFOA *next;
  struct _DELIMITERINFOA *prev;
  char *pDelimiter;
  int nDelimiterLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} DELIMITERINFOA;

typedef struct _DELIMITERINFOW {
  struct _DELIMITERINFOW *next;
  struct _DELIMITERINFOW *prev;
  wchar_t *wpDelimiter;
  int nDelimiterLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} DELIMITERINFOW;

typedef struct _WORDINFOA {
  struct _WORDINFOA *next;
  struct _WORDINFOA *prev;
  char *pWord;
  int nWordLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} WORDINFOA;

typedef struct _WORDINFOW {
  struct _WORDINFOW *next;
  struct _WORDINFOW *prev;
  wchar_t *wpWord;
  int nWordLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} WORDINFOW;

typedef struct _QUOTEINFOA {
  struct _QUOTEINFOA *next;
  struct _QUOTEINFOA *prev;
  char *pQuoteStart;
  int nQuoteStartLen;
  char *pQuoteEnd;
  int nQuoteEndLen;
  char chEscape;
  char *pQuoteInclude;
  int nQuoteIncludeLen;
  char *pQuoteExclude;
  int nQuoteExcludeLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} QUOTEINFOA;

typedef struct _QUOTEINFOW {
  struct _QUOTEINFOW *next;
  struct _QUOTEINFOW *prev;
  wchar_t *wpQuoteStart;
  int nQuoteStartLen;
  wchar_t *wpQuoteEnd;
  int nQuoteEndLen;
  wchar_t wchEscape;
  wchar_t *wpQuoteInclude;
  int nQuoteIncludeLen;
  wchar_t *wpQuoteExclude;
  int nQuoteExcludeLen;
  DWORD dwFlags;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
} QUOTEINFOW;

typedef struct _HEXTENSIONSTACK {
  int first;
  int last;
  int lpExtensionLens[MAX_PATH];
} HEXTENSIONSTACK;

typedef struct _HDELIMSTACK {
  int first;
  int last;
} HDELIMSTACK;

typedef struct _HWORDSTACK {
  int first;
  int last;
  int lpWordLens[MAX_PATH];
} HWORDSTACK;

typedef struct _HQUOTESTACK {
  int first;
  int last;
} HQUOTESTACK;

typedef struct _HSYNTAXFILEA {
  struct _HSYNTAXFILEA *next;
  struct _HSYNTAXFILEA *prev;
  HEXTENSIONSTACK hExtensionStack;
  HDELIMSTACK hDelimiterStack;
  HWORDSTACK hWordStack;
  HQUOTESTACK hQuoteStack;
  char szSyntaxFileName[MAX_PATH];
  HANDLE hThemeHandle;
  DWORD dwFontFlags;
  LOGFONTA lfFont;
  AECOLORS aecColors;
  BOOL bCache;
} HSYNTAXFILEA;

typedef struct _HSYNTAXFILEW {
  struct _HSYNTAXFILEW *next;
  struct _HSYNTAXFILEW *prev;
  HEXTENSIONSTACK hExtensionStack;
  HDELIMSTACK hDelimiterStack;
  HWORDSTACK hWordStack;
  HQUOTESTACK hQuoteStack;
  wchar_t wszSyntaxFileName[MAX_PATH];
  HANDLE hThemeHandle;
  DWORD dwFontFlags;
  LOGFONTW lfFont;
  AECOLORS aecColors;
  BOOL bCache;
} HSYNTAXFILEW;

typedef struct _HWINDOWEDITA {
  struct _HWINDOWEDITA *next;
  struct _HWINDOWEDITA *prev;
  HWND hWnd;
  BOOL bOldFont;
  DWORD dwFontFlags;
  HFONT hFontNew;
  HFONT hFontOld;
  BOOL bOldColors;
  AECOLORS aecNew;
  AECOLORS aecOld;
} HWINDOWEDITA;

typedef struct _HWINDOWEDITW {
  struct _HWINDOWEDITW *next;
  struct _HWINDOWEDITW *prev;
  HWND hWnd;
  BOOL bOldFont;
  DWORD dwFontFlags;
  HFONT hFontNew;
  HFONT hFontOld;
  BOOL bOldColors;
  AECOLORS aecNew;
  AECOLORS aecOld;
} HWINDOWEDITW;

typedef struct _MANUALSETA {
  struct _MANUALSETA *next;
  struct _MANUALSETA *prev;
  HWND hWnd;
  char szExt[MAX_PATH];
} MANUALSETA;

typedef struct _MANUALSETW {
  struct _MANUALSETW *next;
  struct _MANUALSETW *prev;
  HWND hWnd;
  wchar_t wszExt[MAX_PATH];
} MANUALSETW;

typedef struct _MARKTHEME {
  struct _MARKTHEME *next;
  struct _MARKTHEME *prev;
  HANDLE hMarkThemeHandle;
  BOOL bCreated;
  HSTACK hMarkTextsStack;
} MARKTHEME;

typedef struct _MARKTEXT {
  struct _MARKTEXT *next;
  struct _MARKTEXT *prev;
  HANDLE hMarkTextHandle;
  DWORD dwMarkID;
} MARKTEXT;

//Functions prototypes
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
HSYNTAXFILEA* StackLoadSyntaxFileA(char *pFile, HSTACK *hStack, HSYNTAXFILEA *lpSyntaxFile);
HSYNTAXFILEW* StackLoadSyntaxFileW(wchar_t *wpFile, HSTACK *hStack, HSYNTAXFILEW *lpSyntaxFile);
HSYNTAXFILEA* StackInsertSyntaxFileA(HSTACK *hStack);
HSYNTAXFILEW* StackInsertSyntaxFileW(HSTACK *hStack);
HSYNTAXFILEA* StackGetSyntaxFileByExtA(HSTACK *hStack, char *pExtension, int nExtensionLen);
HSYNTAXFILEW* StackGetSyntaxFileByExtW(HSTACK *hStack, wchar_t *wpExtension, int nExtensionLen);
HSYNTAXFILEA* StackGetSyntaxFileByThemeA(HSTACK *hStack, HANDLE hTheme);
HSYNTAXFILEW* StackGetSyntaxFileByThemeW(HSTACK *hStack, HANDLE hTheme);
void StackFreeSyntaxFileA(HSTACK *hStack);
void StackFreeSyntaxFileW(HSTACK *hStack);
EXTENSIONINFOA* StackInsertExtensionA(HEXTENSIONSTACK *hStack, int nExtensionLen);
EXTENSIONINFOW* StackInsertExtensionW(HEXTENSIONSTACK *hStack, int nExtensionLen);
EXTENSIONINFOA* StackGetExtensionA(HEXTENSIONSTACK *hStack, char *pExtension, int nExtensionLen);
EXTENSIONINFOW* StackGetExtensionW(HEXTENSIONSTACK *hStack, wchar_t *wpExtension, int nExtensionLen);
void StackFreeExtensionA(HEXTENSIONSTACK *hStack);
void StackFreeExtensionW(HEXTENSIONSTACK *hStack);
DELIMITERINFOA* StackInsertDelimiterA(HDELIMSTACK *hStack, int nDelimiterLen);
DELIMITERINFOW* StackInsertDelimiterW(HDELIMSTACK *hStack, int nDelimiterLen);
DELIMITERINFOA* StackGetDelimiterA(HDELIMSTACK *hStack, char chDelimiter);
DELIMITERINFOW* StackGetDelimiterW(HDELIMSTACK *hStack, wchar_t wchDelimiter);
void StackFreeDelimiterA(HDELIMSTACK *hStack);
void StackFreeDelimiterW(HDELIMSTACK *hStack);
WORDINFOA* StackInsertWordA(HWORDSTACK *hStack, int nWordLen);
WORDINFOW* StackInsertWordW(HWORDSTACK *hStack, int nWordLen);
WORDINFOA* StackGetWordA(HWORDSTACK *hStack, char *pWord, int nWordLen);
WORDINFOW* StackGetWordW(HWORDSTACK *hStack, wchar_t *wpWord, int nWordLen);
void StackFreeWordA(HWORDSTACK *hStack);
void StackFreeWordW(HWORDSTACK *hStack);
QUOTEINFOA* StackInsertQuoteA(HQUOTESTACK *hStack, int nQuoteStartLen);
QUOTEINFOW* StackInsertQuoteW(HQUOTESTACK *hStack, int nQuoteStartLen);
void StackFreeQuoteA(HQUOTESTACK *hStack);
void StackFreeQuoteW(HQUOTESTACK *hStack);
HWINDOWEDITA* StackInsertWindowA(HSTACK *hStack);
HWINDOWEDITW* StackInsertWindowW(HSTACK *hStack);
HWINDOWEDITA* StackGetWindowA(HSTACK *hStack, HWND hWnd);
HWINDOWEDITW* StackGetWindowW(HSTACK *hStack, HWND hWnd);
void StackDeleteWindowA(HSTACK *hStack, HWINDOWEDITA *lpElement);
void StackDeleteWindowW(HSTACK *hStack, HWINDOWEDITW *lpElement);
void StackFreeWindowA(HSTACK *hStack);
void StackFreeWindowW(HSTACK *hStack);
MANUALSETA* StackInsertManualA(HSTACK *hStack);
MANUALSETW* StackInsertManualW(HSTACK *hStack);
MANUALSETA* StackGetManualA(HSTACK *hStack, HWND hWnd);
MANUALSETW* StackGetManualW(HSTACK *hStack, HWND hWnd);
void StackDeleteManual(HSTACK *hStack, void *ms);
MARKTHEME* StackInsertMarkTheme(HSTACK *hMarkThemesStack);
MARKTHEME* StackGetMarkTheme(HSTACK *hMarkThemesStack, HANDLE hTheme);
void StackDeleteMarkTheme(HSTACK *hMarkThemesStack, MARKTHEME *lpMarkTheme);
void StackFreeMarkTheme(HWND hWnd, HSTACK *hMarkThemesStack);
MARKTEXT* StackInsertMarkText(HSTACK *hMarkTextsStack);
MARKTEXT* StackGetMarkText(HSTACK *hMarkTextsStack, DWORD dwMarkID);
BOOL StackIsAnyMarkText(HSTACK *hMarkTextsStack);
void StackDeleteMarkText(HWND hWnd, MARKTHEME *lpMarkTheme, MARKTEXT *lpMarkText);
void StackFreeMarkText(HWND hWnd, MARKTHEME *lpMarkTheme);
BOOL MarkSelection(HWND hWnd, DWORD dwColorText, DWORD dwColorBk, BOOL bMatchCase, DWORD dwFontStyle, DWORD dwMarkID);
BOOL UnmarkSelection(HWND hWnd, DWORD dwMarkID);
void CreateEditThemeA(HSYNTAXFILEA *lpSyntaxFileA, HWND hWnd);
void CreateEditThemeW(HSYNTAXFILEW *lpSyntaxFileW, HWND hWnd);
void GetFontAndColorsA(HSYNTAXFILEA *lpSyntaxFileA, HWINDOWEDITA *lpWindow);
void GetFontAndColorsW(HSYNTAXFILEW *lpSyntaxFileW, HWINDOWEDITW *lpWindow);
void RestoreFontAndColorsA(HWINDOWEDITA *lpWindow);
void RestoreFontAndColorsW(HWINDOWEDITW *lpWindow);
void UnassignThemeA(HWND hWnd);
void UnassignThemeW(HWND hWnd);
void ClearCache();
void UpdateEdit(HWND hWnd);
void UpdateEditAll();
void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt, TEXTMETRICA *tm);
int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote);
int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote);
BOOL NextLineA(char **pText);
BOOL NextLineW(wchar_t **wpText);
BOOL SkipCommentA(char **pText);
BOOL SkipCommentW(wchar_t **wpText);
int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen);
int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen);
char* GetFileExtA(char *pFile);
wchar_t* GetFileExtW(wchar_t *wpFile);
DWORD GetColorsToRestore(HWND hWnd, AECOLORS *aecHighlight);
COLORREF GetColorValueFromStrA(char *pColor);
COLORREF GetColorValueFromStrW(wchar_t *wpColor);
char* GetColorStrFromValueA(COLORREF crColor, char *szColor);
wchar_t* GetColorStrFromValueW(COLORREF crColor, wchar_t *wszColor);
void ReadSyntaxFilesA();
void ReadSyntaxFilesW();
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hMdiClient;
HICON hMainIcon;
BOOL bOldWindows;
BOOL bOldRichEdit;
BOOL bAkelEdit;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
HSTACK hSyntaxFilesStack={0};
HSTACK hWindowsStack={0};
HSTACK hManualStack={0};
HSTACK hMarkThemeStack={0};
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;
WNDPROCDATA *NewEditProcData=NULL;
char szHighLightDir[MAX_PATH];
wchar_t wszHighLightDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
BOOL bGlobalFontStyleEnable=FALSE;
int nGlobalFontStyle=0;
BOOL bAutoMarkEnable=FALSE;
int nAutoMarkType=MARKAUTO_SYMBOLS;
DWORD dwAutoMarkTextColor=RGB(0x00, 0x00, 0x00);
DWORD dwAutoMarkBkColor=RGB(0x75, 0xD9, 0x75);
BOOL bUseCache=TRUE;
BOOL bSaveOptions=FALSE;
BOOL bUpdateTheme=TRUE;
HWND hWndEditCurrent=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="HighLight";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (pd->lParam)
  {
    if (bInitMain)
    {
      if (IsExtCallParamValid(pd->lParam, 1))
      {
        if (bOldWindows)
        {
          if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_SETEXTENSION)
          {
            MANUALSETA *ms;
            char *pExt=NULL;

            if (IsExtCallParamValid(pd->lParam, 2))
              pExt=(char *)GetExtCallParam(pd->lParam, 2);

            if (pExt)
            {
              if (pd->hWndEdit)
              {
                if (!(ms=StackGetManualA(&hManualStack, pd->hWndEdit)))
                  ms=StackInsertManualA(&hManualStack);

                if (ms)
                {
                  ms->hWnd=pd->hWndEdit;
                  lstrcpynA(ms->szExt, pExt, MAX_PATH);

                  bUpdateTheme=TRUE;
                  UpdateEdit(pd->hWndEdit);
                }
              }
            }
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_MARKSELECTION)
          {
            char *pStringColor1=NULL;
            char *pStringColor2=NULL;
            DWORD dwColor1=(DWORD)-1;
            DWORD dwColor2=(DWORD)-1;
            BOOL bMatchCase=TRUE;
            DWORD dwFontStyle=0;
            DWORD dwMarkID=MARKID_TEXT;
            BOOL bUpdate=FALSE;

            if (IsExtCallParamValid(pd->lParam, 2))
              pStringColor1=(char *)GetExtCallParam(pd->lParam, 2);
            if (IsExtCallParamValid(pd->lParam, 3))
              pStringColor2=(char *)GetExtCallParam(pd->lParam, 3);
            if (IsExtCallParamValid(pd->lParam, 4))
              bMatchCase=GetExtCallParam(pd->lParam, 4);
            if (IsExtCallParamValid(pd->lParam, 5))
              dwFontStyle=GetExtCallParam(pd->lParam, 5);
            if (IsExtCallParamValid(pd->lParam, 6))
              dwMarkID=GetExtCallParam(pd->lParam, 6);

            if (pStringColor1 && *pStringColor1 == '#')
              dwColor1=GetColorValueFromStrA(pStringColor1 + 1);
            if (pStringColor2 && *pStringColor2 == '#')
              dwColor2=GetColorValueFromStrA(pStringColor2 + 1);

            if (dwColor1 != (DWORD)-1 || dwColor2 != (DWORD)-1)
            {
              if (UnmarkSelection(pd->hWndEdit, dwMarkID))
                bUpdate=TRUE;
              if (MarkSelection(pd->hWndEdit, dwColor1, dwColor2, bMatchCase, dwFontStyle, dwMarkID))
                bUpdate=TRUE;
              if (bUpdate)
                UpdateEditAll();
            }
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_UNMARKSELECTION)
          {
            DWORD dwMarkID=0;

            if (IsExtCallParamValid(pd->lParam, 2))
              dwMarkID=GetExtCallParam(pd->lParam, 2);

            if (UnmarkSelection(pd->hWndEdit, dwMarkID))
              UpdateEditAll();
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_CLEARCACHE)
          {
            ClearCache();
          }
        }
        else
        {
          if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_SETEXTENSION)
          {
            MANUALSETW *ms;
            wchar_t *wpExt=NULL;

            if (IsExtCallParamValid(pd->lParam, 2))
              wpExt=(wchar_t *)GetExtCallParam(pd->lParam, 2);

            if (wpExt)
            {
              if (pd->hWndEdit)
              {
                if (!(ms=StackGetManualW(&hManualStack, pd->hWndEdit)))
                  ms=StackInsertManualW(&hManualStack);

                if (ms)
                {
                  ms->hWnd=pd->hWndEdit;
                  lstrcpynW(ms->wszExt, wpExt, MAX_PATH);

                  bUpdateTheme=TRUE;
                  UpdateEdit(pd->hWndEdit);
                }
              }
            }
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_MARKSELECTION)
          {
            wchar_t *wpStringColor1=NULL;
            wchar_t *wpStringColor2=NULL;
            DWORD dwColor1=(DWORD)-1;
            DWORD dwColor2=(DWORD)-1;
            BOOL bMatchCase=TRUE;
            DWORD dwFontStyle=0;
            DWORD dwMarkID=MARKID_TEXT;
            BOOL bUpdate=FALSE;

            if (IsExtCallParamValid(pd->lParam, 2))
              wpStringColor1=(wchar_t *)GetExtCallParam(pd->lParam, 2);
            if (IsExtCallParamValid(pd->lParam, 3))
              wpStringColor2=(wchar_t *)GetExtCallParam(pd->lParam, 3);
            if (IsExtCallParamValid(pd->lParam, 4))
              bMatchCase=GetExtCallParam(pd->lParam, 4);
            if (IsExtCallParamValid(pd->lParam, 5))
              dwFontStyle=GetExtCallParam(pd->lParam, 5);
            if (IsExtCallParamValid(pd->lParam, 6))
              dwMarkID=GetExtCallParam(pd->lParam, 6);

            if (wpStringColor1 && *wpStringColor1 == '#')
              dwColor1=GetColorValueFromStrW(wpStringColor1 + 1);
            if (wpStringColor2 && *wpStringColor2 == '#')
              dwColor2=GetColorValueFromStrW(wpStringColor2 + 1);

            if (dwColor1 != (DWORD)-1 || dwColor2 != (DWORD)-1)
            {
              if (UnmarkSelection(pd->hWndEdit, dwMarkID))
                bUpdate=TRUE;
              if (MarkSelection(pd->hWndEdit, dwColor1, dwColor2, bMatchCase, dwFontStyle, dwMarkID))
                bUpdate=TRUE;
              if (bUpdate)
                UpdateEditAll();
            }
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_UNMARKSELECTION)
          {
            DWORD dwMarkID=0;

            if (IsExtCallParamValid(pd->lParam, 2))
              dwMarkID=GetExtCallParam(pd->lParam, 2);

            if (UnmarkSelection(pd->hWndEdit, dwMarkID))
              UpdateEditAll();
          }
          else if (GetExtCallParam(pd->lParam, 1) == DLLA_HIGHLIGHT_CLEARCACHE)
          {
            ClearCache();
          }
        }
      }

      //Stay in memory, and don't change active status
      pd->nUnload=UD_NONUNLOAD_UNCHANGE;
    }
  }
  else
  {
    if (bInitMain)
    {
      UninitMain();
      UpdateEditAll();
    }
    else
    {
      InitMain();
      if (!pd->bOnStart) UpdateEditAll();

      //Stay in memory, and show as active
      pd->nUnload=UD_NONUNLOAD_ACTIVE;
    }
  }
}

extern "C" void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bOldWindows)
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcA);
  else
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if (pd->bInMemory) pd->nUnload=UD_NONUNLOAD_NONACTIVE;
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndGlobalFontStyleGroup;
  static HWND hWndGlobalFontStyleEnable;
  static HWND hWndGlobalFontStyleIgnore;
  static HWND hWndGlobalFontStyleNormal;
  static HWND hWndGlobalFontStyleBold;
  static HWND hWndGlobalFontStyleItalic;
  static HWND hWndGlobalFontStyleBoldItalic;
  static HWND hWndAutoMarkGroup;
  static HWND hWndAutoMarkEnable;
  static HWND hWndAutoMarkSymbols;
  static HWND hWndAutoMarkWords;
  static HWND hWndAutoMarkTextColorLabel;
  static HWND hWndAutoMarkTextColor;
  static HWND hWndAutoMarkBkColorLabel;
  static HWND hWndAutoMarkBkColor;
  static HWND hWndCacheEnable;
  static HFONT hFontNormal;
  static HFONT hFontBold;
  static HFONT hFontItalic;
  static HFONT hFontBoldItalic;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndGlobalFontStyleGroup=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_GROUP);
    hWndGlobalFontStyleEnable=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_ENABLE);
    hWndGlobalFontStyleIgnore=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_IGNORE);
    hWndGlobalFontStyleNormal=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_NORMAL);
    hWndGlobalFontStyleBold=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_BOLD);
    hWndGlobalFontStyleItalic=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_ITALIC);
    hWndGlobalFontStyleBoldItalic=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_BOLDITALIC);
    hWndAutoMarkGroup=GetDlgItem(hDlg, IDC_AUTOMARK_GROUP);
    hWndAutoMarkEnable=GetDlgItem(hDlg, IDC_AUTOMARK_ENABLE);
    hWndAutoMarkSymbols=GetDlgItem(hDlg, IDC_AUTOMARK_SYMBOLS);
    hWndAutoMarkWords=GetDlgItem(hDlg, IDC_AUTOMARK_WORDS);
    hWndAutoMarkTextColorLabel=GetDlgItem(hDlg, IDC_AUTOMARK_TEXTCOLOR_LABEL);
    hWndAutoMarkTextColor=GetDlgItem(hDlg, IDC_AUTOMARK_TEXTCOLOR);
    hWndAutoMarkBkColorLabel=GetDlgItem(hDlg, IDC_AUTOMARK_BGCOLOR_LABEL);
    hWndAutoMarkBkColor=GetDlgItem(hDlg, IDC_AUTOMARK_BGCOLOR);
    hWndCacheEnable=GetDlgItem(hDlg, IDC_CACHE_ENABLE);

    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_GROUP, GetLangStringA(wLangSystem, STRID_GLOBAL_FONTSTYLE_GROUP));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_ENABLE, GetLangStringA(wLangSystem, STRID_ENABLE));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_IGNORE, GetLangStringA(wLangSystem, STRID_IGNORE));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_NORMAL, GetLangStringA(wLangSystem, STRID_NORMAL));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_BOLD, GetLangStringA(wLangSystem, STRID_BOLD));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_ITALIC, GetLangStringA(wLangSystem, STRID_ITALIC));
    SetDlgItemTextA(hDlg, IDC_GLOBAL_FONTSTYLE_BOLDITALIC, GetLangStringA(wLangSystem, STRID_BOLDITALIC));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_GROUP, GetLangStringA(wLangSystem, STRID_AUTOMARK_GROUP));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_ENABLE, GetLangStringA(wLangSystem, STRID_ENABLE));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_SYMBOLS, GetLangStringA(wLangSystem, STRID_SYMBOLS));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_WORDS, GetLangStringA(wLangSystem, STRID_WORDS));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_TEXTCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_TEXTCOLOR));
    SetDlgItemTextA(hDlg, IDC_AUTOMARK_BGCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_BGCOLOR));
    SetDlgItemTextA(hDlg, IDC_CACHE_ENABLE, GetLangStringA(wLangSystem, STRID_CACHE_ENABLE));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    if (bGlobalFontStyleEnable)
      SendMessage(hWndGlobalFontStyleEnable, BM_SETCHECK, BST_CHECKED, 0);
    if (nGlobalFontStyle == 0)
      SendMessage(hWndGlobalFontStyleIgnore, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 1)
      SendMessage(hWndGlobalFontStyleNormal, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 2)
      SendMessage(hWndGlobalFontStyleBold, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 3)
      SendMessage(hWndGlobalFontStyleItalic, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 4)
      SendMessage(hWndGlobalFontStyleBoldItalic, BM_SETCHECK, BST_CHECKED, 0);
    if (bAutoMarkEnable)
      SendMessage(hWndAutoMarkEnable, BM_SETCHECK, BST_CHECKED, 0);
    if (nAutoMarkType == MARKAUTO_SYMBOLS)
      SendMessage(hWndAutoMarkSymbols, BM_SETCHECK, BST_CHECKED, 0);
    else if (nAutoMarkType == MARKAUTO_WORDS)
      SendMessage(hWndAutoMarkWords, BM_SETCHECK, BST_CHECKED, 0);
    if (bUseCache)
      SendMessage(hWndCacheEnable, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndAutoMarkTextColor, EM_LIMITTEXT, 6, 0);
    SendMessage(hWndAutoMarkBkColor, EM_LIMITTEXT, 6, 0);
    SetWindowTextA(hWndAutoMarkTextColor, GetColorStrFromValueA(dwAutoMarkTextColor, szBuffer));
    SetWindowTextA(hWndAutoMarkBkColor, GetColorStrFromValueA(dwAutoMarkBkColor, szBuffer));

    SendMessage(hDlg, WM_COMMAND, IDC_GLOBAL_FONTSTYLE_ENABLE, 0);
    SendMessage(hDlg, WM_COMMAND, IDC_AUTOMARK_ENABLE, 0);

    //Set fonts
    {
      LOGFONTA lfGui;
      HFONT hGuiFont;

      hGuiFont=(HFONT)GetStockObject(DEFAULT_GUI_FONT);
      GetObjectA((HGDIOBJ)hGuiFont, sizeof(LOGFONTA), &lfGui);

      //Create normal font
      lfGui.lfWeight=FW_NORMAL;
      lfGui.lfItalic=FALSE;
      hFontNormal=(HFONT)CreateFontIndirectA(&lfGui);
      SendMessage(hWndGlobalFontStyleNormal, WM_SETFONT, (WPARAM)hFontNormal, TRUE);

      //Create bold font
      lfGui.lfWeight=FW_BOLD;
      lfGui.lfItalic=FALSE;
      hFontBold=(HFONT)CreateFontIndirectA(&lfGui);
      SendMessage(hWndGlobalFontStyleBold, WM_SETFONT, (WPARAM)hFontBold, TRUE);

      //Create italic font
      lfGui.lfWeight=FW_NORMAL;
      lfGui.lfItalic=TRUE;
      hFontItalic=(HFONT)CreateFontIndirectA(&lfGui);
      SendMessage(hWndGlobalFontStyleItalic, WM_SETFONT, (WPARAM)hFontItalic, TRUE);

      //Create bold italic font
      lfGui.lfWeight=FW_BOLD;
      lfGui.lfItalic=TRUE;
      hFontBoldItalic=(HFONT)CreateFontIndirectA(&lfGui);
      SendMessage(hWndGlobalFontStyleBoldItalic, WM_SETFONT, (WPARAM)hFontBoldItalic, TRUE);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_GLOBAL_FONTSTYLE_ENABLE)
    {
      bGlobalFontStyleEnable=SendMessage(hWndGlobalFontStyleEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndGlobalFontStyleIgnore, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleNormal, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleBold, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleItalic, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleBoldItalic, bGlobalFontStyleEnable);
    }
    else if (LOWORD(wParam) == IDC_AUTOMARK_ENABLE)
    {
      bAutoMarkEnable=SendMessage(hWndAutoMarkEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndAutoMarkSymbols, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkWords, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkTextColorLabel, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkTextColor, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkBkColorLabel, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkBkColor, bAutoMarkEnable);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDCANCEL)
    {
      if (LOWORD(wParam) == IDOK)
      {
        if (SendMessage(hWndGlobalFontStyleIgnore, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=0;
        else if (SendMessage(hWndGlobalFontStyleNormal, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=1;
        else if (SendMessage(hWndGlobalFontStyleBold, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=2;
        else if (SendMessage(hWndGlobalFontStyleItalic, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=3;
        else if (SendMessage(hWndGlobalFontStyleBoldItalic, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=4;

        if (SendMessage(hWndAutoMarkSymbols, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nAutoMarkType=MARKAUTO_SYMBOLS;
        else if (SendMessage(hWndAutoMarkWords, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nAutoMarkType=MARKAUTO_WORDS;
        GetWindowTextA(hWndAutoMarkTextColor, szBuffer, BUFFER_SIZE);
        dwAutoMarkTextColor=GetColorValueFromStrA(szBuffer);
        GetWindowTextA(hWndAutoMarkBkColor, szBuffer, BUFFER_SIZE);
        dwAutoMarkBkColor=GetColorValueFromStrA(szBuffer);

        bUseCache=SendMessage(hWndCacheEnable, BM_GETCHECK, 0, 0);
        SaveOptionsA(OF_SETTINGS);
        ClearCache();
      }
      SendMessage(hWndGlobalFontStyleNormal, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleBold, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleItalic, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleBoldItalic, WM_SETFONT, (WPARAM)NULL, TRUE);
      if (hFontNormal) DeleteObject(hFontNormal);
      if (hFontBold) DeleteObject(hFontBold);
      if (hFontItalic) DeleteObject(hFontItalic);
      if (hFontBoldItalic) DeleteObject(hFontBoldItalic);

      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndGlobalFontStyleGroup;
  static HWND hWndGlobalFontStyleEnable;
  static HWND hWndGlobalFontStyleIgnore;
  static HWND hWndGlobalFontStyleNormal;
  static HWND hWndGlobalFontStyleBold;
  static HWND hWndGlobalFontStyleItalic;
  static HWND hWndGlobalFontStyleBoldItalic;
  static HWND hWndAutoMarkGroup;
  static HWND hWndAutoMarkEnable;
  static HWND hWndAutoMarkSymbols;
  static HWND hWndAutoMarkWords;
  static HWND hWndAutoMarkTextColorLabel;
  static HWND hWndAutoMarkTextColor;
  static HWND hWndAutoMarkBkColorLabel;
  static HWND hWndAutoMarkBkColor;
  static HWND hWndCacheEnable;
  static HFONT hFontNormal;
  static HFONT hFontBold;
  static HFONT hFontItalic;
  static HFONT hFontBoldItalic;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndGlobalFontStyleGroup=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_GROUP);
    hWndGlobalFontStyleEnable=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_ENABLE);
    hWndGlobalFontStyleIgnore=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_IGNORE);
    hWndGlobalFontStyleNormal=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_NORMAL);
    hWndGlobalFontStyleBold=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_BOLD);
    hWndGlobalFontStyleItalic=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_ITALIC);
    hWndGlobalFontStyleBoldItalic=GetDlgItem(hDlg, IDC_GLOBAL_FONTSTYLE_BOLDITALIC);
    hWndAutoMarkGroup=GetDlgItem(hDlg, IDC_AUTOMARK_GROUP);
    hWndAutoMarkEnable=GetDlgItem(hDlg, IDC_AUTOMARK_ENABLE);
    hWndAutoMarkSymbols=GetDlgItem(hDlg, IDC_AUTOMARK_SYMBOLS);
    hWndAutoMarkWords=GetDlgItem(hDlg, IDC_AUTOMARK_WORDS);
    hWndAutoMarkTextColorLabel=GetDlgItem(hDlg, IDC_AUTOMARK_TEXTCOLOR_LABEL);
    hWndAutoMarkTextColor=GetDlgItem(hDlg, IDC_AUTOMARK_TEXTCOLOR);
    hWndAutoMarkBkColorLabel=GetDlgItem(hDlg, IDC_AUTOMARK_BGCOLOR_LABEL);
    hWndAutoMarkBkColor=GetDlgItem(hDlg, IDC_AUTOMARK_BGCOLOR);
    hWndCacheEnable=GetDlgItem(hDlg, IDC_CACHE_ENABLE);

    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_GROUP, GetLangStringW(wLangSystem, STRID_GLOBAL_FONTSTYLE_GROUP));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_ENABLE, GetLangStringW(wLangSystem, STRID_ENABLE));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_IGNORE, GetLangStringW(wLangSystem, STRID_IGNORE));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_NORMAL, GetLangStringW(wLangSystem, STRID_NORMAL));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_BOLD, GetLangStringW(wLangSystem, STRID_BOLD));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_ITALIC, GetLangStringW(wLangSystem, STRID_ITALIC));
    SetDlgItemTextW(hDlg, IDC_GLOBAL_FONTSTYLE_BOLDITALIC, GetLangStringW(wLangSystem, STRID_BOLDITALIC));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_GROUP, GetLangStringW(wLangSystem, STRID_AUTOMARK_GROUP));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_ENABLE, GetLangStringW(wLangSystem, STRID_ENABLE));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_SYMBOLS, GetLangStringW(wLangSystem, STRID_SYMBOLS));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_WORDS, GetLangStringW(wLangSystem, STRID_WORDS));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_TEXTCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_TEXTCOLOR));
    SetDlgItemTextW(hDlg, IDC_AUTOMARK_BGCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_BGCOLOR));
    SetDlgItemTextW(hDlg, IDC_CACHE_ENABLE, GetLangStringW(wLangSystem, STRID_CACHE_ENABLE));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    if (bGlobalFontStyleEnable)
      SendMessage(hWndGlobalFontStyleEnable, BM_SETCHECK, BST_CHECKED, 0);
    if (nGlobalFontStyle == 0)
      SendMessage(hWndGlobalFontStyleIgnore, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 1)
      SendMessage(hWndGlobalFontStyleNormal, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 2)
      SendMessage(hWndGlobalFontStyleBold, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 3)
      SendMessage(hWndGlobalFontStyleItalic, BM_SETCHECK, BST_CHECKED, 0);
    else if (nGlobalFontStyle == 4)
      SendMessage(hWndGlobalFontStyleBoldItalic, BM_SETCHECK, BST_CHECKED, 0);
    if (bAutoMarkEnable)
      SendMessage(hWndAutoMarkEnable, BM_SETCHECK, BST_CHECKED, 0);
    if (nAutoMarkType == MARKAUTO_SYMBOLS)
      SendMessage(hWndAutoMarkSymbols, BM_SETCHECK, BST_CHECKED, 0);
    else if (nAutoMarkType == MARKAUTO_WORDS)
      SendMessage(hWndAutoMarkWords, BM_SETCHECK, BST_CHECKED, 0);
    if (bUseCache)
      SendMessage(hWndCacheEnable, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndAutoMarkTextColor, EM_LIMITTEXT, 6, 0);
    SendMessage(hWndAutoMarkBkColor, EM_LIMITTEXT, 6, 0);
    SetWindowTextW(hWndAutoMarkTextColor, GetColorStrFromValueW(dwAutoMarkTextColor, wszBuffer));
    SetWindowTextW(hWndAutoMarkBkColor, GetColorStrFromValueW(dwAutoMarkBkColor, wszBuffer));

    SendMessage(hDlg, WM_COMMAND, IDC_GLOBAL_FONTSTYLE_ENABLE, 0);
    SendMessage(hDlg, WM_COMMAND, IDC_AUTOMARK_ENABLE, 0);

    //Set fonts
    {
      LOGFONTW lfGui;
      HFONT hGuiFont;

      hGuiFont=(HFONT)GetStockObject(DEFAULT_GUI_FONT);
      GetObjectW((HGDIOBJ)hGuiFont, sizeof(LOGFONTW), &lfGui);

      //Create normal font
      lfGui.lfWeight=FW_NORMAL;
      lfGui.lfItalic=FALSE;
      hFontNormal=(HFONT)CreateFontIndirectW(&lfGui);
      SendMessage(hWndGlobalFontStyleNormal, WM_SETFONT, (WPARAM)hFontNormal, TRUE);

      //Create bold font
      lfGui.lfWeight=FW_BOLD;
      lfGui.lfItalic=FALSE;
      hFontBold=(HFONT)CreateFontIndirectW(&lfGui);
      SendMessage(hWndGlobalFontStyleBold, WM_SETFONT, (WPARAM)hFontBold, TRUE);

      //Create italic font
      lfGui.lfWeight=FW_NORMAL;
      lfGui.lfItalic=TRUE;
      hFontItalic=(HFONT)CreateFontIndirectW(&lfGui);
      SendMessage(hWndGlobalFontStyleItalic, WM_SETFONT, (WPARAM)hFontItalic, TRUE);

      //Create bold italic font
      lfGui.lfWeight=FW_BOLD;
      lfGui.lfItalic=TRUE;
      hFontBoldItalic=(HFONT)CreateFontIndirectW(&lfGui);
      SendMessage(hWndGlobalFontStyleBoldItalic, WM_SETFONT, (WPARAM)hFontBoldItalic, TRUE);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_GLOBAL_FONTSTYLE_ENABLE)
    {
      bGlobalFontStyleEnable=SendMessage(hWndGlobalFontStyleEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndGlobalFontStyleIgnore, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleNormal, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleBold, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleItalic, bGlobalFontStyleEnable);
      EnableWindow(hWndGlobalFontStyleBoldItalic, bGlobalFontStyleEnable);
    }
    else if (LOWORD(wParam) == IDC_AUTOMARK_ENABLE)
    {
      bAutoMarkEnable=SendMessage(hWndAutoMarkEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndAutoMarkSymbols, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkWords, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkTextColorLabel, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkTextColor, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkBkColorLabel, bAutoMarkEnable);
      EnableWindow(hWndAutoMarkBkColor, bAutoMarkEnable);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDCANCEL)
    {
      if (LOWORD(wParam) == IDOK)
      {
        if (SendMessage(hWndGlobalFontStyleIgnore, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=0;
        else if (SendMessage(hWndGlobalFontStyleNormal, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=1;
        else if (SendMessage(hWndGlobalFontStyleBold, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=2;
        else if (SendMessage(hWndGlobalFontStyleItalic, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=3;
        else if (SendMessage(hWndGlobalFontStyleBoldItalic, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nGlobalFontStyle=4;

        if (SendMessage(hWndAutoMarkSymbols, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nAutoMarkType=MARKAUTO_SYMBOLS;
        else if (SendMessage(hWndAutoMarkWords, BM_GETCHECK, 0, 0) == BST_CHECKED)
          nAutoMarkType=MARKAUTO_WORDS;
        GetWindowTextW(hWndAutoMarkTextColor, wszBuffer, BUFFER_SIZE);
        dwAutoMarkTextColor=GetColorValueFromStrW(wszBuffer);
        GetWindowTextW(hWndAutoMarkBkColor, wszBuffer, BUFFER_SIZE);
        dwAutoMarkBkColor=GetColorValueFromStrW(wszBuffer);

        bUseCache=SendMessage(hWndCacheEnable, BM_GETCHECK, 0, 0);
        SaveOptionsW(OF_SETTINGS);
        ClearCache();
      }
      SendMessage(hWndGlobalFontStyleNormal, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleBold, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleItalic, WM_SETFONT, (WPARAM)NULL, TRUE);
      SendMessage(hWndGlobalFontStyleBoldItalic, WM_SETFONT, (WPARAM)NULL, TRUE);
      if (hFontNormal) DeleteObject(hFontNormal);
      if (hFontBold) DeleteObject(hFontBold);
      if (hFontItalic) DeleteObject(hFontItalic);
      if (hFontBoldItalic) DeleteObject(hFontBoldItalic);

      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_OPENDOCUMENT_FINISH ||
      uMsg == AKDN_SAVEDOCUMENT_FINISH)
  {
    bUpdateTheme=TRUE;
    UpdateEdit((HWND)wParam);
  }
  else if (uMsg == AKDN_EDIT_ONFINISH)
  {
    if (bOldWindows)
    {
      MANUALSETA *lpManual;
      HWINDOWEDITA *lpWindow;

      if (lpManual=StackGetManualA(&hManualStack, (HWND)wParam))
        StackDeleteManual(&hManualStack, lpManual);
      if (lpWindow=StackGetWindowA(&hWindowsStack, (HWND)wParam))
        StackDeleteWindowA(&hWindowsStack, lpWindow);
      UnmarkSelection((HWND)wParam, 0);
    }
    else
    {
      MANUALSETW *lpManual;
      HWINDOWEDITW *lpWindow;

      if (lpManual=StackGetManualW(&hManualStack, (HWND)wParam))
        StackDeleteManual(&hManualStack, lpManual);
      if (lpWindow=StackGetWindowW(&hWindowsStack, (HWND)wParam))
        StackDeleteWindowW(&hWindowsStack, lpWindow);
      UnmarkSelection((HWND)wParam, 0);
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == AEN_SELCHANGED)
      {
        if (bAutoMarkEnable)
        {
          AENSELCHANGE *aensc=(AENSELCHANGE *)lParam;
          CHARRANGE cr;

          SendMessage(aensc->hdr.hwndFrom, EM_EXGETSEL, 0, (LPARAM)&cr);

          if (cr.cpMax - cr.cpMin > 0 && cr.cpMax - cr.cpMin < MARKMAX_WORD &&
              nAutoMarkType == MARKAUTO_SYMBOLS ||
              (nAutoMarkType == MARKAUTO_WORDS &&
               SendMessage(aensc->hdr.hwndFrom, AEM_ISDELIMITER, AEDLM_WORD|AEDLM_PREVCHAR, (LPARAM)&aensc->aes.crSel.ciMin) &&
               !SendMessage(aensc->hdr.hwndFrom, AEM_ISDELIMITER, AEDLM_WORD, (LPARAM)&aensc->aes.crSel.ciMin) &&
               SendMessage(aensc->hdr.hwndFrom, AEM_ISDELIMITER, AEDLM_WORD, (LPARAM)&aensc->aes.crSel.ciMax) &&
               !SendMessage(aensc->hdr.hwndFrom, AEM_ISDELIMITER, AEDLM_WORD|AEDLM_PREVCHAR, (LPARAM)&aensc->aes.crSel.ciMax)))
          {
            UnmarkSelection(aensc->hdr.hwndFrom, MARKID_AUTO);
            MarkSelection(aensc->hdr.hwndFrom, dwAutoMarkTextColor, dwAutoMarkBkColor, 1, 0, MARKID_AUTO);
            UpdateEdit(aensc->hdr.hwndFrom);
          }
          else
          {
            if (UnmarkSelection(aensc->hdr.hwndFrom, MARKID_AUTO))
              UpdateEdit(aensc->hdr.hwndFrom);
          }
        }
      }
    }
  }

  //Call next procedure
  if (hWnd == hMainWnd)
    return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_PAINT)
  {
    static HSYNTAXFILEA *lpSyntaxFileA=NULL;

    //Check scheme
    if (bUpdateTheme || hWnd != hWndEditCurrent)
    {
      MANUALSETA *lpManual;
      HWINDOWEDITA *lpWindow;
      EDITINFO ei;
      char *pExt=NULL;

      if (!bOldRichEdit)
      {
        EDITINFO ei;

        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
        {
          if (hWnd != ei.hWndEdit)
            return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
        }
      }
      bUpdateTheme=FALSE;
      hWndEditCurrent=hWnd;

      if (lpManual=StackGetManualA(&hManualStack, hWnd))
      {
        pExt=lpManual->szExt;
      }
      else
      {
        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
          pExt=GetFileExtA((char *)ei.pFile);
      }

      if (pExt)
      {
        if (lpSyntaxFileA=StackGetSyntaxFileByExtA(&hSyntaxFilesStack, pExt, lstrlenA(pExt)))
        {
          if (lpSyntaxFileA->bCache)
          {
            char szHighLightFile[MAX_PATH];

            lpSyntaxFileA->bCache=FALSE;
            StackFreeExtensionA(&lpSyntaxFileA->hExtensionStack);
            wsprintfA(szHighLightFile, "%s\\%s.highlight", szHighLightDir, lpSyntaxFileA->szSyntaxFileName);
            StackLoadSyntaxFileA(szHighLightFile, &hSyntaxFilesStack, lpSyntaxFileA);
            bSaveOptions=TRUE;
          }
          if (bAkelEdit)
          {
            if (!lpSyntaxFileA->hThemeHandle)
              CreateEditThemeA(lpSyntaxFileA, hWnd);

            if (lpSyntaxFileA->hThemeHandle != (HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEA, 0, (LPARAM)NULL))
            {
              if (!(lpWindow=StackGetWindowA(&hWindowsStack, hWnd)))
              {
                if (lpWindow=StackInsertWindowA(&hWindowsStack))
                  lpWindow->hWnd=hWnd;
              }

              if (lpWindow)
              {
                //Assign theme
                UnmarkSelection(hWnd, 0);
                SendMessage(hWnd, AEM_HLSETTHEME, (WPARAM)lpSyntaxFileA->hThemeHandle, FALSE);

                //Set colors and font
                RestoreFontAndColorsA(lpWindow);
                GetFontAndColorsA(lpSyntaxFileA, lpWindow);

                if (lpWindow->aecNew.dwFlags)
                  SendMessage(hWnd, AEM_SETCOLORS, 0, (LPARAM)&lpWindow->aecNew);
                if (lpWindow->hFontNew)
                  SendMessage(hWnd, WM_SETFONT, (WPARAM)lpWindow->hFontNew, TRUE);
              }
            }
          }
        }
        else
        {
          //Extension doesn't associated, remove all highlights
          if (bAkelEdit)
          {
            if (lpWindow=StackGetWindowA(&hWindowsStack, hWnd))
              RestoreFontAndColorsA(lpWindow);
            UnassignThemeA(hWnd);
          }
        }
      }
    }

    //HighLight RichEdit (AkelPad 3.x.x)
    if (!bAkelEdit && lpSyntaxFileA)
    {
      GETTEXTRANGE gtr;
      CHARRANGE cr;
      TEXTMETRICA tm;
      RECT rcUpdate;
      RECT rcEdit;
      RECT rcDraw;
      HDC hDC;
      HRGN hRgn;
      HRGN hRgnOld;
      HFONT hFont;
      HFONT hFontOld;
      LRESULT lResult;
      int nFirstChar;
      int nLastChar;
      int nFirstCharInUpdateLine;
      int i;

      GetUpdateRect(hWnd, &rcUpdate, FALSE);
      SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
      if (rcUpdate.left < rcEdit.left) rcUpdate.left=rcEdit.left;
      if (rcUpdate.top < rcEdit.top) rcUpdate.top=rcEdit.top;
      if (rcUpdate.right > rcEdit.right) rcUpdate.right=rcEdit.right;
      if (rcUpdate.bottom > rcEdit.bottom) rcUpdate.bottom=rcEdit.bottom;
      rcDraw.left=rcEdit.left;
      rcDraw.top=rcUpdate.top;
      rcDraw.right=rcEdit.right;
      rcDraw.bottom=rcUpdate.bottom;

      lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

      if (hDC=GetDC(hWnd))
      {
        HideCaret(hWnd);
        hRgn=CreateRectRgn(rcEdit.left, rcEdit.top, rcEdit.right, rcEdit.bottom);
        hRgnOld=(HRGN)SelectObject(hDC, hRgn);
        hFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
        hFontOld=(HFONT)SelectObject(hDC, hFont);
        GetTextMetricsA(hDC, &tm);
        SetBkMode(hDC, TRANSPARENT);

        i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.left);
        i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
        nFirstChar=SendMessage(hWnd, EM_LINEINDEX, i, 0);
        i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.right);
        i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
        i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
        nLastChar=i + SendMessage(hWnd, EM_LINELENGTH, i, 0);
        gtr.cpMin=nFirstChar;
        gtr.cpMax=nLastChar;
        SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);

        if (SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr))
        {
          DELIMITERINFOA *lpDelimElement=NULL;
          WORDINFOA *lpWordElement=NULL;
          char szWord[MAX_PATH];
          char *szText=(char *)gtr.pText;
          char *pStrBegin;
          char *pStrEnd;
          POINT ptUpdateLine;
          POINT ptDelimiter;
          POINT ptWordBegin;
          POINT ptWordEnd;
          POINT ptSelBegin;
          POINT ptSelEnd;
          int nDelimiterIndex;
          int nWordBeginIndex;
          int nWordEndIndex;
          int nWordLen;

          //Get first visible word
          nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcUpdate);
          if (nFirstCharInUpdateLine > nFirstChar) --nFirstCharInUpdateLine;
          if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
          pStrEnd=szText + (nFirstCharInUpdateLine - nFirstChar);

          while (pStrEnd > szText && (*pStrEnd != ' ' && *pStrEnd != '\t' && *pStrEnd != '\r' && !StackGetDelimiterA(&lpSyntaxFileA->hDelimiterStack, *pStrEnd)))
            --pStrEnd;
          while (*pStrEnd == ' ' || *pStrEnd == '\t' || *pStrEnd == '\r')
            ++pStrEnd;
          pStrBegin=pStrEnd;

          GetPosFromChar(hWnd, nFirstCharInUpdateLine, &ptUpdateLine, &tm);
          ptUpdateLine.x=rcUpdate.left;

          while (1)
          {
            lpDelimElement=NULL;

            if (*pStrEnd == ' ' ||
                *pStrEnd == '\t' ||
                *pStrEnd == '\r' ||
                *pStrEnd == '\0' ||
                (lpDelimElement=StackGetDelimiterA(&lpSyntaxFileA->hDelimiterStack, *pStrEnd)))
            {
              ////DELIMITER
              if (lpDelimElement)
              {
                if ((int)lpDelimElement->dwColor2 != -1 || (int)lpDelimElement->dwColor1 != -1)
                {
                  nDelimiterIndex=nFirstChar + (pStrEnd - szText);
                  GetPosFromChar(hWnd, nDelimiterIndex, &ptDelimiter, &tm);
                  if (ptDelimiter.y < ptUpdateLine.y) goto FreeText;

                  if (ptDelimiter.y > ptUpdateLine.y ||
                      (pStrBegin == pStrEnd &&
                       ptDelimiter.y == ptUpdateLine.y &&
                       ptDelimiter.x >= rcUpdate.right &&
                       nDelimiterIndex > nFirstCharInUpdateLine))
                  {
                    ptUpdateLine.y+=tm.tmHeight;

                    nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                    i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                    i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                    if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                    if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                    pStrBegin=szText + (i - nFirstChar);
                    pStrEnd=szText + (nFirstCharInUpdateLine - nFirstChar);

                    while (pStrEnd > pStrBegin && (*pStrEnd != ' ' && *pStrEnd != '\t' && *pStrEnd != '\r' && !StackGetDelimiterA(&lpSyntaxFileA->hDelimiterStack, *pStrEnd)))
                      --pStrEnd;
                    while (*pStrEnd == ' ' || *pStrEnd == '\t' || *pStrEnd == '\r')
                      ++pStrEnd;
                    pStrBegin=pStrEnd;
                    continue;
                  }
                  else
                  {
                    if (ptDelimiter.x + tm.tmMaxCharWidth > rcUpdate.left)
                    {
                      rcDraw.left=ptDelimiter.x;
                      rcDraw.top=ptDelimiter.y;

                      if (!(nDelimiterIndex >= cr.cpMin && nDelimiterIndex < cr.cpMax))
                      {
                        if ((int)lpDelimElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpDelimElement->dwColor1);
                          DrawTextA(hDC, pStrEnd, 1, &rcDraw, DT_NOPREFIX);
                        }
                      }
                      else
                      {
                        if ((int)lpDelimElement->dwColor2 != -1)
                        {
                          SetTextColor(hDC, lpDelimElement->dwColor2);
                          DrawTextA(hDC, pStrEnd, 1, &rcDraw, DT_NOPREFIX);
                        }
                      }
                    }
                  }
                }
              }
              nWordLen=pStrEnd - pStrBegin;

              ////WORD
              if (nWordLen && nWordLen < MAX_PATH)
              {
                nWordBeginIndex=nFirstChar + (pStrBegin - szText);
                nWordEndIndex=nFirstChar + (pStrEnd - szText);
                GetPosFromChar(hWnd, nWordBeginIndex, &ptWordBegin, &tm);
                GetPosFromChar(hWnd, nWordEndIndex, &ptWordEnd, &tm);
                if (ptWordBegin.y < ptUpdateLine.y) goto FreeText;

                if (ptWordBegin.y == ptWordEnd.y)
                {
                  if (bOldRichEdit)
                  {
                    //RichEdit 2.0: make word highlited, even if its beginning is not displayed
                    if (ptWordBegin.x > ptWordEnd.x) ptWordBegin.x-=0x10000;
                  }
                }

                if (ptWordBegin.y != ptWordEnd.y ||
                    ptWordBegin.y > ptUpdateLine.y ||
                    (ptWordBegin.y == ptUpdateLine.y &&
                     ptWordBegin.x >= rcUpdate.right &&
                     nWordBeginIndex > nFirstCharInUpdateLine))
                {
                  ptUpdateLine.y+=tm.tmHeight;

                  nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                  i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                  i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                  if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                  if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                  pStrBegin=szText + (i - nFirstChar);
                  pStrEnd=szText + (nFirstCharInUpdateLine - nFirstChar);

                  while (pStrEnd > pStrBegin && (*pStrEnd != ' ' && *pStrEnd != '\t' && *pStrEnd != '\r' && !StackGetDelimiterA(&lpSyntaxFileA->hDelimiterStack, *pStrEnd)))
                    --pStrEnd;
                  while (*pStrEnd == ' ' || *pStrEnd == '\t' || *pStrEnd == '\r')
                    ++pStrEnd;
                  pStrBegin=pStrEnd;
                  continue;
                }
                else
                {
                  if (ptWordEnd.x > rcUpdate.left)
                  {
                    lstrcpynA(szWord, pStrBegin, nWordLen + 1);
                    rcDraw.left=ptWordBegin.x;
                    rcDraw.top=ptWordBegin.y;

                    if (lpWordElement=StackGetWordA(&lpSyntaxFileA->hWordStack, szWord, nWordLen))
                    {
                      if (nWordBeginIndex >= cr.cpMax || nWordEndIndex <= cr.cpMin)
                      {
                        if ((int)lpWordElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpWordElement->dwColor1);
                          DrawTextA(hDC, pStrBegin, nWordLen, &rcDraw, DT_NOPREFIX);
                        }
                      }
                      else
                      {
                        //Highlight selection
                        GetPosFromChar(hWnd, cr.cpMin, &ptSelBegin, &tm);
                        GetPosFromChar(hWnd, cr.cpMax, &ptSelEnd, &tm);

                        if ((int)lpWordElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpWordElement->dwColor1);

                          if (nWordBeginIndex < cr.cpMin)
                          {
                            rcDraw.left=ptWordBegin.x;
                            DrawTextA(hDC, pStrBegin, cr.cpMin - nWordBeginIndex, &rcDraw, DT_NOPREFIX);
                          }
                          if (nWordEndIndex > cr.cpMax)
                          {
                            rcDraw.left=ptSelEnd.x;
                            DrawTextA(hDC, pStrBegin + (cr.cpMax - nWordBeginIndex), nWordEndIndex - cr.cpMax, &rcDraw, DT_NOPREFIX);
                          }
                        }
                        if ((int)lpWordElement->dwColor2 != -1)
                        {
                          if (cr.cpMin > nWordBeginIndex)
                            rcDraw.left=ptSelBegin.x;
                          else
                            rcDraw.left=ptWordBegin.x;
                          SetTextColor(hDC, lpWordElement->dwColor2);
                          DrawTextA(hDC, pStrBegin + (max(cr.cpMin, nWordBeginIndex) - nWordBeginIndex), min(cr.cpMax, nWordEndIndex) - max(cr.cpMin, nWordBeginIndex), &rcDraw, DT_NOPREFIX);
                        }
                      }
                    }
                  }
                }
              }
              if (!*pStrEnd) break;
              while (*++pStrEnd == ' ' || *pStrEnd == '\t' || *pStrEnd == '\r');
              pStrBegin=pStrEnd;
            }
            else ++pStrEnd;
          }
          FreeText:
          SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
        }
        if (hFontOld) SelectObject(hDC, hFontOld);
        DeleteObject(hFont);
        if (hRgnOld) SelectObject(hDC, hRgnOld);
        DeleteObject(hRgn);
        ShowCaret(hWnd);
        ReleaseDC(hWnd, hDC);
      }
      return lResult;
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_PAINT)
  {
    static HSYNTAXFILEW *lpSyntaxFileW=NULL;

    //Check scheme
    if (bUpdateTheme || hWnd != hWndEditCurrent)
    {
      MANUALSETW *lpManual;
      HWINDOWEDITW *lpWindow;
      EDITINFO ei;
      wchar_t *wpExt=NULL;

      bUpdateTheme=FALSE;
      hWndEditCurrent=hWnd;

      if (lpManual=StackGetManualW(&hManualStack, hWnd))
      {
        wpExt=lpManual->wszExt;
      }
      else
      {
        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
          wpExt=GetFileExtW((wchar_t *)ei.pFile);
      }
      if (wpExt)
      {
        if (lpSyntaxFileW=StackGetSyntaxFileByExtW(&hSyntaxFilesStack, wpExt, lstrlenW(wpExt)))
        {
          if (lpSyntaxFileW->bCache)
          {
            wchar_t wszHighLightFile[MAX_PATH];

            lpSyntaxFileW->bCache=FALSE;
            StackFreeExtensionW(&lpSyntaxFileW->hExtensionStack);
            wsprintfW(wszHighLightFile, L"%s\\%s.highlight", wszHighLightDir, lpSyntaxFileW->wszSyntaxFileName);
            StackLoadSyntaxFileW(wszHighLightFile, &hSyntaxFilesStack, lpSyntaxFileW);
            bSaveOptions=TRUE;
          }
          if (bAkelEdit)
          {
            if (!lpSyntaxFileW->hThemeHandle)
              CreateEditThemeW(lpSyntaxFileW, hWnd);

            if (lpSyntaxFileW->hThemeHandle != (HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEW, 0, (LPARAM)NULL))
            {
              if (!(lpWindow=StackGetWindowW(&hWindowsStack, hWnd)))
              {
                if (lpWindow=StackInsertWindowW(&hWindowsStack))
                  lpWindow->hWnd=hWnd;
              }

              if (lpWindow)
              {
                //Assign theme
                UnmarkSelection(hWnd, 0);
                SendMessage(hWnd, AEM_HLSETTHEME, (WPARAM)lpSyntaxFileW->hThemeHandle, FALSE);

                //Set colors and font
                RestoreFontAndColorsW(lpWindow);
                GetFontAndColorsW(lpSyntaxFileW, lpWindow);

                if (lpWindow->aecNew.dwFlags)
                  SendMessage(hWnd, AEM_SETCOLORS, 0, (LPARAM)&lpWindow->aecNew);
                if (lpWindow->hFontNew)
                  SendMessage(hWnd, WM_SETFONT, (WPARAM)lpWindow->hFontNew, TRUE);
              }
            }
          }
        }
        else
        {
          //Extension doesn't associated, remove all highlights
          if (bAkelEdit)
          {
            if (lpWindow=StackGetWindowW(&hWindowsStack, hWnd))
              RestoreFontAndColorsW(lpWindow);
            UnassignThemeW(hWnd);
          }
        }
      }
    }

    //HighLight RichEdit (AkelPad 3.x.x)
    if (!bAkelEdit && lpSyntaxFileW)
    {
      GETTEXTRANGE gtr;
      CHARRANGE cr;
      TEXTMETRICW tm;
      RECT rcUpdate;
      RECT rcEdit;
      RECT rcDraw;
      HDC hDC;
      HRGN hRgn;
      HRGN hRgnOld;
      HFONT hFont;
      HFONT hFontOld;
      LRESULT lResult;
      int nFirstChar;
      int nLastChar;
      int nFirstCharInUpdateLine;
      int i;

      GetUpdateRect(hWnd, &rcUpdate, FALSE);
      SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
      if (rcUpdate.left < rcEdit.left) rcUpdate.left=rcEdit.left;
      if (rcUpdate.top < rcEdit.top) rcUpdate.top=rcEdit.top;
      if (rcUpdate.right > rcEdit.right) rcUpdate.right=rcEdit.right;
      if (rcUpdate.bottom > rcEdit.bottom) rcUpdate.bottom=rcEdit.bottom;
      rcDraw.left=rcEdit.left;
      rcDraw.top=rcUpdate.top;
      rcDraw.right=rcEdit.right;
      rcDraw.bottom=rcUpdate.bottom;

      lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

      if (hDC=GetDC(hWnd))
      {
        HideCaret(hWnd);
        hRgn=CreateRectRgn(rcEdit.left, rcEdit.top, rcEdit.right, rcEdit.bottom);
        hRgnOld=(HRGN)SelectObject(hDC, hRgn);
        hFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
        hFontOld=(HFONT)SelectObject(hDC, hFont);
        GetTextMetricsW(hDC, &tm);
        SetBkMode(hDC, TRANSPARENT);

        i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.left);
        i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
        nFirstChar=SendMessage(hWnd, EM_LINEINDEX, i, 0);
        i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.right);
        i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
        i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
        nLastChar=i + SendMessage(hWnd, EM_LINELENGTH, i, 0);
        gtr.cpMin=nFirstChar;
        gtr.cpMax=nLastChar;
        SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);

        if (SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr))
        {
          DELIMITERINFOW *lpDelimElement=NULL;
          WORDINFOW *lpWordElement=NULL;
          wchar_t wszWord[MAX_PATH];
          wchar_t *wszText=(wchar_t *)gtr.pText;
          wchar_t *wpStrBegin;
          wchar_t *wpStrEnd;
          POINT ptUpdateLine;
          POINT ptDelimiter;
          POINT ptWordBegin;
          POINT ptWordEnd;
          POINT ptSelBegin;
          POINT ptSelEnd;
          int nDelimiterIndex;
          int nWordBeginIndex;
          int nWordEndIndex;
          int nWordLen;

          //Get first visible word
          nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcUpdate);
          if (nFirstCharInUpdateLine > nFirstChar) --nFirstCharInUpdateLine;
          if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
          wpStrEnd=wszText + (nFirstCharInUpdateLine - nFirstChar);

          while (wpStrEnd > wszText && (*wpStrEnd != ' ' && *wpStrEnd != '\t' && *wpStrEnd != '\r' && !StackGetDelimiterW(&lpSyntaxFileW->hDelimiterStack, *wpStrEnd)))
            --wpStrEnd;
          while (*wpStrEnd == ' ' || *wpStrEnd == '\t' || *wpStrEnd == '\r')
            ++wpStrEnd;
          wpStrBegin=wpStrEnd;

          GetPosFromChar(hWnd, nFirstCharInUpdateLine, &ptUpdateLine, (TEXTMETRICA *)&tm);
          ptUpdateLine.x=rcUpdate.left;

          while (1)
          {
            lpDelimElement=NULL;

            if (*wpStrEnd == ' ' ||
                *wpStrEnd == '\t' ||
                *wpStrEnd == '\r' ||
                *wpStrEnd == '\0' ||
                (lpDelimElement=StackGetDelimiterW(&lpSyntaxFileW->hDelimiterStack, *wpStrEnd)))
            {
              ////DELIMITER
              if (lpDelimElement)
              {
                if ((int)lpDelimElement->dwColor2 != -1 || (int)lpDelimElement->dwColor1 != -1)
                {
                  nDelimiterIndex=nFirstChar + (wpStrEnd - wszText);
                  GetPosFromChar(hWnd, nDelimiterIndex, &ptDelimiter, (TEXTMETRICA *)&tm);
                  if (ptDelimiter.y < ptUpdateLine.y) goto FreeText;

                  if (ptDelimiter.y > ptUpdateLine.y ||
                      (wpStrBegin == wpStrEnd &&
                       ptDelimiter.y == ptUpdateLine.y &&
                       ptDelimiter.x >= rcUpdate.right &&
                       nDelimiterIndex > nFirstCharInUpdateLine))
                  {
                    ptUpdateLine.y+=tm.tmHeight;

                    nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                    i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                    i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                    if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                    if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                    wpStrBegin=wszText + (i - nFirstChar);
                    wpStrEnd=wszText + (nFirstCharInUpdateLine - nFirstChar);

                    while (wpStrEnd > wpStrBegin && (*wpStrEnd != ' ' && *wpStrEnd != '\t' && *wpStrEnd != '\r' && !StackGetDelimiterW(&lpSyntaxFileW->hDelimiterStack, *wpStrEnd)))
                      --wpStrEnd;
                    while (*wpStrEnd == ' ' || *wpStrEnd == '\t' || *wpStrEnd == '\r')
                      ++wpStrEnd;
                    wpStrBegin=wpStrEnd;
                    continue;
                  }
                  else
                  {
                    if (ptDelimiter.x + tm.tmMaxCharWidth > rcUpdate.left)
                    {
                      rcDraw.left=ptDelimiter.x;
                      rcDraw.top=ptDelimiter.y;

                      if (!(nDelimiterIndex >= cr.cpMin && nDelimiterIndex < cr.cpMax))
                      {
                        if ((int)lpDelimElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpDelimElement->dwColor1);
                          DrawTextW(hDC, wpStrEnd, 1, &rcDraw, DT_NOPREFIX);
                        }
                      }
                      else
                      {
                        if ((int)lpDelimElement->dwColor2 != -1)
                        {
                          SetTextColor(hDC, lpDelimElement->dwColor2);
                          DrawTextW(hDC, wpStrEnd, 1, &rcDraw, DT_NOPREFIX);
                        }
                      }
                    }
                  }
                }
              }
              nWordLen=wpStrEnd - wpStrBegin;

              ////WORD
              if (nWordLen && nWordLen < MAX_PATH)
              {
                nWordBeginIndex=nFirstChar + (wpStrBegin - wszText);
                nWordEndIndex=nFirstChar + (wpStrEnd - wszText);
                GetPosFromChar(hWnd, nWordBeginIndex, &ptWordBegin, (TEXTMETRICA *)&tm);
                GetPosFromChar(hWnd, nWordEndIndex, &ptWordEnd, (TEXTMETRICA *)&tm);
                if (ptWordBegin.y < ptUpdateLine.y) goto FreeText;

                if (ptWordBegin.y == ptWordEnd.y)
                {
                  if (bOldRichEdit)
                  {
                    //RichEdit 2.0: make word highlited, even if its beginning is not displayed
                    if (ptWordBegin.x > ptWordEnd.x) ptWordBegin.x-=0x10000;
                  }
                }

                if (ptWordBegin.y != ptWordEnd.y ||
                    ptWordBegin.y > ptUpdateLine.y ||
                    (ptWordBegin.y == ptUpdateLine.y &&
                     ptWordBegin.x >= rcUpdate.right &&
                     nWordBeginIndex > nFirstCharInUpdateLine))
                {
                  ptUpdateLine.y+=tm.tmHeight;

                  nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                  i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                  i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                  if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                  if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                  wpStrBegin=wszText + (i - nFirstChar);
                  wpStrEnd=wszText + (nFirstCharInUpdateLine - nFirstChar);

                  while (wpStrEnd > wpStrBegin && (*wpStrEnd != ' ' && *wpStrEnd != '\t' && *wpStrEnd != '\r' && !StackGetDelimiterW(&lpSyntaxFileW->hDelimiterStack, *wpStrEnd)))
                    --wpStrEnd;
                  while (*wpStrEnd == ' ' || *wpStrEnd == '\t' || *wpStrEnd == '\r')
                    ++wpStrEnd;
                  wpStrBegin=wpStrEnd;
                  continue;
                }
                else
                {
                  if (ptWordEnd.x > rcUpdate.left)
                  {
                    lstrcpynW(wszWord, wpStrBegin, nWordLen + 1);
                    rcDraw.left=ptWordBegin.x;
                    rcDraw.top=ptWordBegin.y;

                    if (lpWordElement=StackGetWordW(&lpSyntaxFileW->hWordStack, wszWord, nWordLen))
                    {
                      if (nWordBeginIndex >= cr.cpMax || nWordEndIndex <= cr.cpMin)
                      {
                        if ((int)lpWordElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpWordElement->dwColor1);
                          DrawTextW(hDC, wpStrBegin, nWordLen, &rcDraw, DT_NOPREFIX);
                        }
                      }
                      else
                      {
                        //Highlight selection
                        GetPosFromChar(hWnd, cr.cpMin, &ptSelBegin, (TEXTMETRICA *)&tm);
                        GetPosFromChar(hWnd, cr.cpMax, &ptSelEnd, (TEXTMETRICA *)&tm);

                        if ((int)lpWordElement->dwColor1 != -1)
                        {
                          SetTextColor(hDC, lpWordElement->dwColor1);

                          if (nWordBeginIndex < cr.cpMin)
                          {
                            rcDraw.left=ptWordBegin.x;
                            DrawTextW(hDC, wpStrBegin, cr.cpMin - nWordBeginIndex, &rcDraw, DT_NOPREFIX);
                          }
                          if (nWordEndIndex > cr.cpMax)
                          {
                            rcDraw.left=ptSelEnd.x;
                            DrawTextW(hDC, wpStrBegin + (cr.cpMax - nWordBeginIndex), nWordEndIndex - cr.cpMax, &rcDraw, DT_NOPREFIX);
                          }
                        }
                        if ((int)lpWordElement->dwColor2 != -1)
                        {
                          if (cr.cpMin > nWordBeginIndex)
                            rcDraw.left=ptSelBegin.x;
                          else
                            rcDraw.left=ptWordBegin.x;
                          SetTextColor(hDC, lpWordElement->dwColor2);
                          DrawTextW(hDC, wpStrBegin + (max(cr.cpMin, nWordBeginIndex) - nWordBeginIndex), min(cr.cpMax, nWordEndIndex) - max(cr.cpMin, nWordBeginIndex), &rcDraw, DT_NOPREFIX);
                        }
                      }
                    }
                  }
                }
              }
              if (!*wpStrEnd) break;
              while (*++wpStrEnd == ' ' || *wpStrEnd == '\t' || *wpStrEnd == '\r');
              wpStrBegin=wpStrEnd;
            }
            else ++wpStrEnd;
          }
          FreeText:
          SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
        }
        if (hFontOld) SelectObject(hDC, hFontOld);
        DeleteObject(hFont);
        if (hRgnOld) SelectObject(hDC, hRgnOld);
        DeleteObject(hRgn);
        ShowCaret(hWnd);
        ReleaseDC(hWnd, hDC);
      }
      return lResult;
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

HSYNTAXFILEA* StackLoadSyntaxFileA(char *pFile, HSTACK *hStack, HSYNTAXFILEA *lpSyntaxFile)
{
  EXTENSIONINFOA *lpExtElement=NULL;
  DELIMITERINFOA *lpDelimElement=NULL;
  WORDINFOA *lpWordElement=NULL;
  QUOTEINFOA *lpQuoteElement=NULL;
  wchar_t *wszText;
  wchar_t *wpText;
  char *szText;
  char *pText;
  char *pExtension;
  char *pDelimiter;
  char *pWord;
  char *pQuoteStart;
  char *pQuoteEnd;
  char *pQuoteInclude;
  char *pQuoteExclude;
  HANDLE hFile;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  DWORD dwAnsiLen;
  int nExtensionLen;
  int nDelimiterLen;
  int nWordLen;
  int nQuoteStartLen;
  int nQuoteEndLen;
  int nQuoteIncludeLen;
  int nQuoteExcludeLen;
  char chEscape;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
  DWORD dwFlags;
  BOOL bQuoteString;

  if ((hFile=CreateFileA(pFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
  {
    if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
    {
      if (dwFileSize >= 2)
      {
        if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
        {
          if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
          {
            dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
            wszText[dwUnicodeLen++]='\0';
            if (wszText[0] == 0xFEFF)
            {
              ++wpText;
              --dwUnicodeLen;
            }

            dwAnsiLen=WideCharToMultiByte(CP_ACP, 0, wpText, dwUnicodeLen, NULL, 0, NULL, NULL);

            if (szText=pText=(char *)GlobalAlloc(GPTR, dwAnsiLen))
            {
              WideCharToMultiByte(CP_ACP, 0, wpText, dwUnicodeLen, pText, dwAnsiLen, NULL, NULL);

              if (!lpSyntaxFile)
              {
                if (lpSyntaxFile=StackInsertSyntaxFileA(hStack))
                  GetBaseNameA(pFile, lpSyntaxFile->szSyntaxFileName, MAX_PATH);
              }
              if (lpSyntaxFile)
              {
                while (1)
                {
                  SectionStart:
                  SkipCommentA(&pText);
                  GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL);
                  if (!NextLineA(&pText)) goto FreeText;

                  if (!lstrcmpiA(szBuffer, "Extensions:"))
                  {
                    while (1)
                    {
                      //Parse line
                      while (1)
                      {
                        lpExtElement=NULL;
                        pExtension=NULL;

                        //Extension
                        if ((nExtensionLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
                        {
                          if (!bQuoteString && szBuffer[0] == ';') break;

                          if (pExtension=(char *)GlobalAlloc(GPTR, nExtensionLen + 1))
                            memcpy(pExtension, szBuffer, nExtensionLen + 1);
                          else
                            break;
                        }
                        else goto SectionStart;

                        //Add to stack
                        if (lpExtElement=StackInsertExtensionA(&lpSyntaxFile->hExtensionStack, nExtensionLen))
                        {
                          lpExtElement->pExtension=pExtension;
                          lpExtElement->nExtensionLen=nExtensionLen;
                        }
                        break;
                      }
                      if (!lpExtElement && pExtension) GlobalFree((HGLOBAL)pExtension);
                      if (!NextLineA(&pText)) goto FreeText;
                    }
                  }
                  else if (!lstrcmpiA(szBuffer, "Font:"))
                  {
                    if (!SkipCommentA(&pText)) goto FreeText;

                    //Parse line
                    while (1)
                    {
                      HDC hDC;
                      int nFontPointSize=0;

                      lpSyntaxFile->dwFontFlags=0;
                      lpSyntaxFile->lfFont.lfHeight=0;
                      lpSyntaxFile->lfFont.lfWidth=0;
                      lpSyntaxFile->lfFont.lfEscapement=0;
                      lpSyntaxFile->lfFont.lfOrientation=0;
                      lpSyntaxFile->lfFont.lfWeight=0;
                      lpSyntaxFile->lfFont.lfItalic=FALSE;
                      lpSyntaxFile->lfFont.lfUnderline=FALSE;
                      lpSyntaxFile->lfFont.lfStrikeOut=FALSE;
                      lpSyntaxFile->lfFont.lfCharSet=DEFAULT_CHARSET;
                      lpSyntaxFile->lfFont.lfOutPrecision=OUT_DEFAULT_PRECIS;
                      lpSyntaxFile->lfFont.lfClipPrecision=CLIP_DEFAULT_PRECIS;
                      lpSyntaxFile->lfFont.lfQuality=DEFAULT_QUALITY;
                      lpSyntaxFile->lfFont.lfPitchAndFamily=DEFAULT_PITCH;
                      lpSyntaxFile->lfFont.lfFaceName[0]='\0';

                      //Font style
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        dwFontStyle=xatoiA(szBuffer);
                        if (dwFontStyle)
                        {
                          lpSyntaxFile->dwFontFlags|=FF_FONTSTYLE;
                          lpSyntaxFile->lfFont.lfWeight=(dwFontStyle == AEHLS_FONTBOLD || dwFontStyle == AEHLS_FONTBOLDITALIC)?FW_BOLD:FW_NORMAL;
                          lpSyntaxFile->lfFont.lfItalic=(dwFontStyle == AEHLS_FONTITALIC || dwFontStyle == AEHLS_FONTBOLDITALIC)?TRUE:FALSE;
                        }
                        else if (szBuffer[0] != '0') break;
                      }
                      else goto SectionStart;

                      //Font size
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        nFontPointSize=xatoiA(szBuffer);
                        if (nFontPointSize)
                        {
                          if (hDC=GetDC(hMainWnd))
                          {
                            lpSyntaxFile->dwFontFlags|=FF_FONTSIZE;
                            lpSyntaxFile->lfFont.lfHeight=-MulDiv(nFontPointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
                            ReleaseDC(hMainWnd, hDC);
                          }
                        }
                        else if (szBuffer[0] != '0') break;
                      }
                      else break;

                      //Face name
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        lpSyntaxFile->dwFontFlags|=FF_FONTFACE;
                        lstrcpynA(lpSyntaxFile->lfFont.lfFaceName, szBuffer, LF_FACESIZE);
                      }
                      else break;

                      break;
                    }
                    goto SectionStart;
                  }
                  else if (!lstrcmpiA(szBuffer, "Colors:"))
                  {
                    if (!SkipCommentA(&pText)) goto FreeText;

                    //Parse line
                    while (1)
                    {
                      lpSyntaxFile->aecColors.dwFlags=0;

                      //Basic text color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crBasicText=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_BASICTEXT;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else goto SectionStart;

                      //Basic bkgrnd color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crBasicBk=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_BASICBK;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Sel text color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crSelText=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_SELTEXT;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Sel bkgrnd color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crSelBk=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_SELBK;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Line text color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crActiveLineText=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVELINETEXT;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Line bkgrnd color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crActiveLineBk=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVELINEBK;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Column color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crActiveColumn=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVECOLUMN;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Marker color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crColumnMarker=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_COLUMNMARKER;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //Caret color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crCaret=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_CARET;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      //URL color
                      if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                      {
                        if (*szBuffer == '#')
                        {
                          lpSyntaxFile->aecColors.crUrlText=GetColorValueFromStrA(szBuffer + 1);
                          lpSyntaxFile->aecColors.dwFlags|=AECLR_URLTEXT;
                        }
                        else if (*szBuffer != '0') break;
                      }
                      else break;

                      break;
                    }
                    goto SectionStart;
                  }
                  else if (!lstrcmpiA(szBuffer, "Quotes:"))
                  {
                    while (1)
                    {
                      //Parse line
                      while (1)
                      {
                        lpQuoteElement=NULL;
                        pQuoteStart=NULL;
                        pQuoteEnd=NULL;
                        pQuoteInclude=NULL;
                        pQuoteExclude=NULL;
                        dwFlags=AEHLF_MATCHCASE;
                        dwColor1=(DWORD)-1;
                        dwColor2=(DWORD)-1;
                        dwFontStyle=0;

                        //Highlight flags
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          dwFlags=xatoiA(szBuffer);
                          if (dwFlags == 0 && szBuffer[0] != '0') break;
                        }
                        else goto SectionStart;

                        //Font style
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                          dwFontStyle=xatoiA(szBuffer);
                        else
                          break;

                        //Color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor1=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Background color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor2=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Quote start
                        if ((nQuoteStartLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
                        {
                          if (pQuoteStart=(char *)GlobalAlloc(GPTR, nQuoteStartLen + 1))
                            memcpy(pQuoteStart, szBuffer, nQuoteStartLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Quote end
                        if ((nQuoteEndLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
                        {
                          if (pQuoteEnd=(char *)GlobalAlloc(GPTR, nQuoteEndLen + 1))
                            memcpy(pQuoteEnd, szBuffer, nQuoteEndLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Escape
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString) || bQuoteString)
                          chEscape=szBuffer[0];
                        else
                          break;

                        //Quote include
                        if ((nQuoteIncludeLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
                        {
                          if (pQuoteInclude=(char *)GlobalAlloc(GPTR, nQuoteIncludeLen + 1))
                            memcpy(pQuoteInclude, szBuffer, nQuoteIncludeLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Quote exclude
                        if ((nQuoteExcludeLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
                        {
                          if (pQuoteExclude=(char *)GlobalAlloc(GPTR, nQuoteExcludeLen + 1))
                            memcpy(pQuoteExclude, szBuffer, nQuoteExcludeLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Add to stack
                        if (lpQuoteElement=StackInsertQuoteA(&lpSyntaxFile->hQuoteStack, nQuoteStartLen))
                        {
                          lpQuoteElement->pQuoteStart=pQuoteStart;
                          lpQuoteElement->nQuoteStartLen=nQuoteStartLen;
                          lpQuoteElement->pQuoteEnd=pQuoteEnd;
                          lpQuoteElement->nQuoteEndLen=nQuoteEndLen;
                          lpQuoteElement->chEscape=chEscape;
                          lpQuoteElement->pQuoteInclude=pQuoteInclude;
                          lpQuoteElement->nQuoteIncludeLen=nQuoteIncludeLen;
                          lpQuoteElement->pQuoteExclude=pQuoteExclude;
                          lpQuoteElement->nQuoteExcludeLen=nQuoteExcludeLen;
                          lpQuoteElement->dwFlags=dwFlags;
                          lpQuoteElement->dwColor1=dwColor1;
                          lpQuoteElement->dwColor2=dwColor2;
                          lpQuoteElement->dwFontStyle=dwFontStyle;
                        }
                        break;
                      }
                      if (!lpQuoteElement)
                      {
                        if (pQuoteStart) GlobalFree((HGLOBAL)pQuoteStart);
                        if (pQuoteEnd) GlobalFree((HGLOBAL)pQuoteEnd);
                        if (pQuoteInclude) GlobalFree((HGLOBAL)pQuoteInclude);
                        if (pQuoteExclude) GlobalFree((HGLOBAL)pQuoteExclude);
                      }
                      if (!NextLineA(&pText)) goto FreeText;
                    }
                  }
                  else if (!lstrcmpiA(szBuffer, "Delimiters:"))
                  {
                    while (1)
                    {
                      //Parse line
                      while (1)
                      {
                        lpDelimElement=NULL;
                        pDelimiter=NULL;
                        dwFlags=AEHLF_MATCHCASE;
                        dwColor1=(DWORD)-1;
                        dwColor2=(DWORD)-1;
                        dwFontStyle=0;

                        //Highlight flags
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          dwFlags=xatoiA(szBuffer);
                          if (dwFlags == 0 && szBuffer[0] != '0') break;
                        }
                        else goto SectionStart;

                        //Font style
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                          dwFontStyle=xatoiA(szBuffer);
                        else
                          break;

                        //Color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor1=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Background color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor2=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Delimiter
                        if (nDelimiterLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (pDelimiter=(char *)GlobalAlloc(GPTR, nDelimiterLen + 1))
                            memcpy(pDelimiter, szBuffer, nDelimiterLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Add to stack
                        if (lpDelimElement=StackInsertDelimiterA(&lpSyntaxFile->hDelimiterStack, nDelimiterLen))
                        {
                          lpDelimElement->pDelimiter=pDelimiter;
                          lpDelimElement->nDelimiterLen=nDelimiterLen;
                          lpDelimElement->dwFlags=dwFlags;
                          lpDelimElement->dwColor1=dwColor1;
                          lpDelimElement->dwColor2=dwColor2;
                          lpDelimElement->dwFontStyle=dwFontStyle;
                        }
                        break;
                      }
                      if (!lpDelimElement && pDelimiter) GlobalFree((HGLOBAL)pDelimiter);
                      if (!NextLineA(&pText)) goto FreeText;
                    }
                  }
                  else if (!lstrcmpiA(szBuffer, "Words:"))
                  {
                    while (1)
                    {
                      //Parse line
                      while (1)
                      {
                        lpWordElement=NULL;
                        pWord=NULL;
                        dwFlags=AEHLF_MATCHCASE;
                        dwColor1=(DWORD)-1;
                        dwColor2=(DWORD)-1;
                        dwFontStyle=0;

                        //Highlight flags
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          dwFlags=xatoiA(szBuffer);
                          if (dwFlags == 0 && szBuffer[0] != '0') break;
                        }
                        else goto SectionStart;

                        //Font style
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                          dwFontStyle=xatoiA(szBuffer);
                        else
                          break;

                        //Color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor1=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Background color
                        if (GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (*szBuffer == '#')
                          {
                            dwColor2=GetColorValueFromStrA(szBuffer + 1);
                          }
                          else if (*szBuffer != '0') break;
                        }
                        else break;

                        //Word
                        if (nWordLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, NULL))
                        {
                          if (pWord=(char *)GlobalAlloc(GPTR, nWordLen + 1))
                            memcpy(pWord, szBuffer, nWordLen + 1);
                          else
                            break;
                        }
                        else break;

                        //Add to stack
                        if (lpWordElement=StackInsertWordA(&lpSyntaxFile->hWordStack, nWordLen))
                        {
                          lpWordElement->pWord=pWord;
                          lpWordElement->nWordLen=nWordLen;
                          lpWordElement->dwFlags=dwFlags;
                          lpWordElement->dwColor1=dwColor1;
                          lpWordElement->dwColor2=dwColor2;
                          lpWordElement->dwFontStyle=dwFontStyle;
                        }
                        break;
                      }
                      if (!lpWordElement && pWord) GlobalFree((HGLOBAL)pWord);
                      if (!NextLineA(&pText)) goto FreeText;
                    }
                  }
                  else break;
                }
              }
              FreeText:
              GlobalFree((HGLOBAL)szText);
            }
          }
          GlobalFree((HGLOBAL)wszText);
        }
      }
    }
    CloseHandle(hFile);
  }
  return lpSyntaxFile;
}

HSYNTAXFILEW* StackLoadSyntaxFileW(wchar_t *wpFile, HSTACK *hStack, HSYNTAXFILEW *lpSyntaxFile)
{
  EXTENSIONINFOW *lpExtElement=NULL;
  DELIMITERINFOW *lpDelimElement=NULL;
  WORDINFOW *lpWordElement=NULL;
  QUOTEINFOW *lpQuoteElement=NULL;
  wchar_t *wszText;
  wchar_t *wpText;
  wchar_t *wpExtension;
  wchar_t *wpDelimiter;
  wchar_t *wpWord;
  wchar_t *wpQuoteStart;
  wchar_t *wpQuoteEnd;
  wchar_t *wpQuoteInclude;
  wchar_t *wpQuoteExclude;
  HANDLE hFile;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  int nExtensionLen;
  int nDelimiterLen;
  int nWordLen;
  int nQuoteStartLen;
  int nQuoteEndLen;
  int nQuoteIncludeLen;
  int nQuoteExcludeLen;
  wchar_t wchEscape;
  DWORD dwColor1;
  DWORD dwColor2;
  DWORD dwFontStyle;
  DWORD dwFlags;
  BOOL bQuoteString;

  if ((hFile=CreateFileW(wpFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
  {
    if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
    {
      if (dwFileSize >= 2)
      {
        if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
        {
          if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
          {
            dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
            wszText[dwUnicodeLen++]='\0';
            if (wszText[0] == 0xFEFF)
            {
              ++wpText;
              --dwUnicodeLen;
            }

            if (!lpSyntaxFile)
            {
              if (lpSyntaxFile=StackInsertSyntaxFileW(hStack))
                GetBaseNameW(wpFile, lpSyntaxFile->wszSyntaxFileName, MAX_PATH);
            }
            if (lpSyntaxFile)
            {
              while (1)
              {
                SectionStart:
                SkipCommentW(&wpText);
                GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL);
                if (!NextLineW(&wpText)) goto FreeText;

                if (!lstrcmpiW(wszBuffer, L"Extensions:"))
                {
                  while (1)
                  {
                    //Parse line
                    while (1)
                    {
                      lpExtElement=NULL;
                      wpExtension=NULL;

                      //Extension
                      if ((nExtensionLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                      {
                        if (!bQuoteString && wszBuffer[0] == ';') break;

                        if (wpExtension=(wchar_t *)GlobalAlloc(GPTR, nExtensionLen * sizeof(wchar_t) + 2))
                          memcpy(wpExtension, wszBuffer, nExtensionLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else goto SectionStart;

                      //Add to stack
                      if (lpExtElement=StackInsertExtensionW(&lpSyntaxFile->hExtensionStack, nExtensionLen))
                      {
                        lpExtElement->wpExtension=wpExtension;
                        lpExtElement->nExtensionLen=nExtensionLen;
                      }
                      break;
                    }
                    if (!lpExtElement && wpExtension) GlobalFree((HGLOBAL)wpExtension);
                    if (!NextLineW(&wpText)) goto FreeText;
                  }
                }
                else if (!lstrcmpiW(wszBuffer, L"Font:"))
                {
                  if (!SkipCommentW(&wpText)) goto FreeText;

                  //Parse line
                  while (1)
                  {
                    HDC hDC;
                    int nFontPointSize=0;

                    lpSyntaxFile->dwFontFlags=0;
                    lpSyntaxFile->lfFont.lfHeight=0;
                    lpSyntaxFile->lfFont.lfWidth=0;
                    lpSyntaxFile->lfFont.lfEscapement=0;
                    lpSyntaxFile->lfFont.lfOrientation=0;
                    lpSyntaxFile->lfFont.lfWeight=0;
                    lpSyntaxFile->lfFont.lfItalic=FALSE;
                    lpSyntaxFile->lfFont.lfUnderline=FALSE;
                    lpSyntaxFile->lfFont.lfStrikeOut=FALSE;
                    lpSyntaxFile->lfFont.lfCharSet=DEFAULT_CHARSET;
                    lpSyntaxFile->lfFont.lfOutPrecision=OUT_DEFAULT_PRECIS;
                    lpSyntaxFile->lfFont.lfClipPrecision=CLIP_DEFAULT_PRECIS;
                    lpSyntaxFile->lfFont.lfQuality=DEFAULT_QUALITY;
                    lpSyntaxFile->lfFont.lfPitchAndFamily=DEFAULT_PITCH;
                    lpSyntaxFile->lfFont.lfFaceName[0]='\0';

                    //Font style
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      dwFontStyle=xatoiW(wszBuffer);
                      if (dwFontStyle)
                      {
                        lpSyntaxFile->dwFontFlags|=FF_FONTSTYLE;
                        lpSyntaxFile->lfFont.lfWeight=(dwFontStyle == AEHLS_FONTBOLD || dwFontStyle == AEHLS_FONTBOLDITALIC)?FW_BOLD:FW_NORMAL;
                        lpSyntaxFile->lfFont.lfItalic=(dwFontStyle == AEHLS_FONTITALIC || dwFontStyle == AEHLS_FONTBOLDITALIC)?TRUE:FALSE;
                      }
                      else if (wszBuffer[0] != '0') break;
                    }
                    else goto SectionStart;

                    //Font size
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      nFontPointSize=xatoiW(wszBuffer);
                      if (nFontPointSize)
                      {
                        if (hDC=GetDC(hMainWnd))
                        {
                          lpSyntaxFile->dwFontFlags|=FF_FONTSIZE;
                          lpSyntaxFile->lfFont.lfHeight=-MulDiv(nFontPointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
                          ReleaseDC(hMainWnd, hDC);
                        }
                      }
                      else if (wszBuffer[0] != '0') break;
                    }
                    else break;

                    //Face name
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      lpSyntaxFile->dwFontFlags|=FF_FONTFACE;
                      lstrcpynW(lpSyntaxFile->lfFont.lfFaceName, wszBuffer, LF_FACESIZE);
                    }
                    else break;

                    break;
                  }
                  goto SectionStart;
                }
                else if (!lstrcmpiW(wszBuffer, L"Colors:"))
                {
                  if (!SkipCommentW(&wpText)) goto FreeText;

                  //Parse line
                  while (1)
                  {
                    lpSyntaxFile->aecColors.dwFlags=0;

                    //Basic text color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crBasicText=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_BASICTEXT;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else goto SectionStart;

                    //Basic bkgrnd color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crBasicBk=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_BASICBK;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Sel text color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crSelText=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_SELTEXT;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Sel bkgrnd color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crSelBk=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_SELBK;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Line text color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crActiveLineText=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVELINETEXT;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Line bkgrnd color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crActiveLineBk=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVELINEBK;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Column color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crActiveColumn=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_ACTIVECOLUMN;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Marker color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crColumnMarker=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_COLUMNMARKER;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //Caret color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crCaret=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_CARET;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    //URL color
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      if (*wszBuffer == '#')
                      {
                        lpSyntaxFile->aecColors.crUrlText=GetColorValueFromStrW(wszBuffer + 1);
                        lpSyntaxFile->aecColors.dwFlags|=AECLR_URLTEXT;
                      }
                      else if (*wszBuffer != '0') break;
                    }
                    else break;

                    break;
                  }
                  goto SectionStart;
                }
                else if (!lstrcmpiW(wszBuffer, L"Quotes:"))
                {
                  while (1)
                  {
                    //Parse line
                    while (1)
                    {
                      lpQuoteElement=NULL;
                      wpQuoteStart=NULL;
                      wpQuoteEnd=NULL;
                      wpQuoteInclude=NULL;
                      wpQuoteExclude=NULL;
                      dwFlags=AEHLF_MATCHCASE;
                      dwColor1=(DWORD)-1;
                      dwColor2=(DWORD)-1;
                      dwFontStyle=0;

                      //Highlight flags
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        dwFlags=xatoiW(wszBuffer);
                        if (dwFlags == 0 && wszBuffer[0] != '0') break;
                      }
                      else goto SectionStart;

                      //Font style
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                        dwFontStyle=xatoiW(wszBuffer);
                      else
                        break;

                      //Color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor1=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Background color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor2=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Quote start
                      if ((nQuoteStartLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                      {
                        if (wpQuoteStart=(wchar_t *)GlobalAlloc(GPTR, nQuoteStartLen * sizeof(wchar_t) + 2))
                          memcpy(wpQuoteStart, wszBuffer, nQuoteStartLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Quote end
                      if ((nQuoteEndLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                      {
                        if (wpQuoteEnd=(wchar_t *)GlobalAlloc(GPTR, nQuoteEndLen * sizeof(wchar_t) + 2))
                          memcpy(wpQuoteEnd, wszBuffer, nQuoteEndLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Escape
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString) || bQuoteString)
                        wchEscape=wszBuffer[0];
                      else
                        break;

                      //Quote include
                      if ((nQuoteIncludeLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                      {
                        if (wpQuoteInclude=(wchar_t *)GlobalAlloc(GPTR, nQuoteIncludeLen * sizeof(wchar_t) + 2))
                          memcpy(wpQuoteInclude, wszBuffer, nQuoteIncludeLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Quote exclude
                      if ((nQuoteExcludeLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                      {
                        if (wpQuoteExclude=(wchar_t *)GlobalAlloc(GPTR, nQuoteExcludeLen * sizeof(wchar_t) + 2))
                          memcpy(wpQuoteExclude, wszBuffer, nQuoteExcludeLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Add to stack
                      if (lpQuoteElement=StackInsertQuoteW(&lpSyntaxFile->hQuoteStack, nQuoteStartLen))
                      {
                        lpQuoteElement->wpQuoteStart=wpQuoteStart;
                        lpQuoteElement->nQuoteStartLen=nQuoteStartLen;
                        lpQuoteElement->wpQuoteEnd=wpQuoteEnd;
                        lpQuoteElement->nQuoteEndLen=nQuoteEndLen;
                        lpQuoteElement->wchEscape=wchEscape;
                        lpQuoteElement->wpQuoteInclude=wpQuoteInclude;
                        lpQuoteElement->nQuoteIncludeLen=nQuoteIncludeLen;
                        lpQuoteElement->wpQuoteExclude=wpQuoteExclude;
                        lpQuoteElement->nQuoteExcludeLen=nQuoteExcludeLen;
                        lpQuoteElement->dwFlags=dwFlags;
                        lpQuoteElement->dwColor1=dwColor1;
                        lpQuoteElement->dwColor2=dwColor2;
                        lpQuoteElement->dwFontStyle=dwFontStyle;
                      }
                      break;
                    }
                    if (!lpQuoteElement)
                    {
                      if (wpQuoteStart) GlobalFree((HGLOBAL)wpQuoteStart);
                      if (wpQuoteEnd) GlobalFree((HGLOBAL)wpQuoteEnd);
                      if (wpQuoteInclude) GlobalFree((HGLOBAL)wpQuoteInclude);
                      if (wpQuoteExclude) GlobalFree((HGLOBAL)wpQuoteExclude);
                    }
                    if (!NextLineW(&wpText)) goto FreeText;
                  }
                }
                else if (!lstrcmpiW(wszBuffer, L"Delimiters:"))
                {
                  while (1)
                  {
                    //Parse line
                    while (1)
                    {
                      lpDelimElement=NULL;
                      wpDelimiter=NULL;
                      dwFlags=AEHLF_MATCHCASE;
                      dwColor1=(DWORD)-1;
                      dwColor2=(DWORD)-1;
                      dwFontStyle=0;

                      //Highlight flags
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        dwFlags=xatoiW(wszBuffer);
                        if (dwFlags == 0 && wszBuffer[0] != '0') break;
                      }
                      else goto SectionStart;

                      //Font style
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                        dwFontStyle=xatoiW(wszBuffer);
                      else
                        break;

                      //Color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor1=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Background color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor2=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Delimiter
                      if (nDelimiterLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (wpDelimiter=(wchar_t *)GlobalAlloc(GPTR, nDelimiterLen * sizeof(wchar_t) + 2))
                          memcpy(wpDelimiter, wszBuffer, nDelimiterLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Add to stack
                      if (lpDelimElement=StackInsertDelimiterW(&lpSyntaxFile->hDelimiterStack, nDelimiterLen))
                      {
                        lpDelimElement->wpDelimiter=wpDelimiter;
                        lpDelimElement->nDelimiterLen=nDelimiterLen;
                        lpDelimElement->dwFlags=dwFlags;
                        lpDelimElement->dwColor1=dwColor1;
                        lpDelimElement->dwColor2=dwColor2;
                        lpDelimElement->dwFontStyle=dwFontStyle;
                      }
                      break;
                    }
                    if (!lpDelimElement && wpDelimiter) GlobalFree((HGLOBAL)wpDelimiter);
                    if (!NextLineW(&wpText)) goto FreeText;
                  }
                }
                else if (!lstrcmpiW(wszBuffer, L"Words:"))
                {
                  while (1)
                  {
                    //Parse line
                    while (1)
                    {
                      lpWordElement=NULL;
                      wpWord=NULL;
                      dwFlags=AEHLF_MATCHCASE;
                      dwColor1=(DWORD)-1;
                      dwColor2=(DWORD)-1;
                      dwFontStyle=0;

                      //Highlight flags
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        dwFlags=xatoiW(wszBuffer);
                        if (dwFlags == 0 && wszBuffer[0] != '0') break;
                      }
                      else goto SectionStart;

                      //Font style
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                        dwFontStyle=xatoiW(wszBuffer);
                      else
                        break;

                      //Color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor1=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Background color
                      if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (*wszBuffer == '#')
                        {
                          dwColor2=GetColorValueFromStrW(wszBuffer + 1);
                        }
                        else if (*wszBuffer != '0') break;
                      }
                      else break;

                      //Word
                      if (nWordLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                      {
                        if (wpWord=(wchar_t *)GlobalAlloc(GPTR, nWordLen * sizeof(wchar_t) + 2))
                          memcpy(wpWord, wszBuffer, nWordLen * sizeof(wchar_t) + 2);
                        else
                          break;
                      }
                      else break;

                      //Add to stack
                      if (lpWordElement=StackInsertWordW(&lpSyntaxFile->hWordStack, nWordLen))
                      {
                        lpWordElement->wpWord=wpWord;
                        lpWordElement->nWordLen=nWordLen;
                        lpWordElement->dwFlags=dwFlags;
                        lpWordElement->dwColor1=dwColor1;
                        lpWordElement->dwColor2=dwColor2;
                        lpWordElement->dwFontStyle=dwFontStyle;
                      }
                      break;
                    }
                    if (!lpWordElement && wpWord) GlobalFree((HGLOBAL)wpWord);
                    if (!NextLineW(&wpText)) goto FreeText;
                  }
                }
                else break;
              }
            }
          }
          FreeText:
          GlobalFree((HGLOBAL)wszText);
        }
      }
    }
    CloseHandle(hFile);
  }
  return lpSyntaxFile;
}

HSYNTAXFILEA* StackInsertSyntaxFileA(HSTACK *hStack)
{
  HSYNTAXFILEA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HSYNTAXFILEA));
  return lpElement;
}

HSYNTAXFILEW* StackInsertSyntaxFileW(HSTACK *hStack)
{
  HSYNTAXFILEW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HSYNTAXFILEW));
  return lpElement;
}

HSYNTAXFILEA* StackGetSyntaxFileByExtA(HSTACK *hStack, char *pExtension, int nExtensionLen)
{
  HSYNTAXFILEA *lpElement=(HSYNTAXFILEA *)hStack->first;

  while (lpElement)
  {
    if (StackGetExtensionA(&lpElement->hExtensionStack, pExtension, nExtensionLen))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HSYNTAXFILEW* StackGetSyntaxFileByExtW(HSTACK *hStack, wchar_t *wpExtension, int nExtensionLen)
{
  HSYNTAXFILEW *lpElement=(HSYNTAXFILEW *)hStack->first;

  while (lpElement)
  {
    if (StackGetExtensionW(&lpElement->hExtensionStack, wpExtension, nExtensionLen))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HSYNTAXFILEA* StackGetSyntaxFileByThemeA(HSTACK *hStack, HANDLE hTheme)
{
  HSYNTAXFILEA *lpElement=(HSYNTAXFILEA *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hThemeHandle == hTheme)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HSYNTAXFILEW* StackGetSyntaxFileByThemeW(HSTACK *hStack, HANDLE hTheme)
{
  HSYNTAXFILEW *lpElement=(HSYNTAXFILEW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hThemeHandle == hTheme)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeSyntaxFileA(HSTACK *hStack)
{
  HSYNTAXFILEA *lpElement=(HSYNTAXFILEA *)hStack->first;
  EDITINFO ei={0};

  if (bAkelEdit)
  {
    SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei);
  }

  while (lpElement)
  {
    StackFreeExtensionA(&lpElement->hExtensionStack);
    StackFreeDelimiterA(&lpElement->hDelimiterStack);
    StackFreeWordA(&lpElement->hWordStack);
    StackFreeQuoteA(&lpElement->hQuoteStack);
    if (ei.hWndEdit && lpElement->hThemeHandle)
      SendMessage(ei.hWndEdit, AEM_HLDELETETHEME, (WPARAM)lpElement->hThemeHandle, 0);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackFreeSyntaxFileW(HSTACK *hStack)
{
  HSYNTAXFILEW *lpElement=(HSYNTAXFILEW *)hStack->first;
  EDITINFO ei={0};

  if (bAkelEdit)
  {
    SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei);
  }

  while (lpElement)
  {
    StackFreeExtensionW(&lpElement->hExtensionStack);
    StackFreeDelimiterW(&lpElement->hDelimiterStack);
    StackFreeWordW(&lpElement->hWordStack);
    StackFreeQuoteW(&lpElement->hQuoteStack);
    if (ei.hWndEdit && lpElement->hThemeHandle)
      SendMessage(ei.hWndEdit, AEM_HLDELETETHEME, (WPARAM)lpElement->hThemeHandle, 0);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

EXTENSIONINFOA* StackInsertExtensionA(HEXTENSIONSTACK *hStack, int nExtensionLen)
{
  EXTENSIONINFOA *lpElement1;
  EXTENSIONINFOA *lpElement2=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    if (hStack->lpExtensionLens[nExtensionLen])
    {
      lpElement1=(EXTENSIONINFOA *)hStack->lpExtensionLens[nExtensionLen];
    }
    else
    {
      lpElement1=(EXTENSIONINFOA *)hStack->first;

      while (lpElement1)
      {
        if (lpElement1->nExtensionLen >= nExtensionLen)
          break;

        lpElement1=lpElement1->next;
      }
    }
    StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(EXTENSIONINFOA));

    if (lpElement2)
      hStack->lpExtensionLens[nExtensionLen]=(int)lpElement2;
  }
  return lpElement2;
}

EXTENSIONINFOW* StackInsertExtensionW(HEXTENSIONSTACK *hStack, int nExtensionLen)
{
  EXTENSIONINFOW *lpElement1;
  EXTENSIONINFOW *lpElement2=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    if (hStack->lpExtensionLens[nExtensionLen])
    {
      lpElement1=(EXTENSIONINFOW *)hStack->lpExtensionLens[nExtensionLen];
    }
    else
    {
      lpElement1=(EXTENSIONINFOW *)hStack->first;

      while (lpElement1)
      {
        if (lpElement1->nExtensionLen >= nExtensionLen)
          break;

        lpElement1=lpElement1->next;
      }
    }
    StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(EXTENSIONINFOW));

    if (lpElement2)
      hStack->lpExtensionLens[nExtensionLen]=(int)lpElement2;
  }
  return lpElement2;
}

EXTENSIONINFOA* StackGetExtensionA(HEXTENSIONSTACK *hStack, char *pExtension, int nExtensionLen)
{
  EXTENSIONINFOA *lpElement=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    lpElement=(EXTENSIONINFOA *)hStack->lpExtensionLens[nExtensionLen];

    while (lpElement)
    {
      if (lpElement->nExtensionLen != nExtensionLen)
        break;
      if (!lstrcmpiA(lpElement->pExtension, pExtension))
        return lpElement;

      lpElement=lpElement->next;
    }
  }
  return NULL;
}

EXTENSIONINFOW* StackGetExtensionW(HEXTENSIONSTACK *hStack, wchar_t *wpExtension, int nExtensionLen)
{
  EXTENSIONINFOW *lpElement=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    lpElement=(EXTENSIONINFOW *)hStack->lpExtensionLens[nExtensionLen];

    while (lpElement)
    {
      if (lpElement->nExtensionLen != nExtensionLen)
        break;
      if (!lstrcmpiW(lpElement->wpExtension, wpExtension))
        return lpElement;

      lpElement=lpElement->next;
    }
  }
  return NULL;
}

void StackFreeExtensionA(HEXTENSIONSTACK *hStack)
{
  EXTENSIONINFOA *lpElement=(EXTENSIONINFOA *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->pExtension);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
  memset(hStack->lpExtensionLens, 0, MAX_PATH * sizeof(int));
}

void StackFreeExtensionW(HEXTENSIONSTACK *hStack)
{
  EXTENSIONINFOW *lpElement=(EXTENSIONINFOW *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpExtension);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
  memset(hStack->lpExtensionLens, 0, MAX_PATH * sizeof(int));
}

DELIMITERINFOA* StackInsertDelimiterA(HDELIMSTACK *hStack, int nDelimiterLen)
{
  DELIMITERINFOA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(DELIMITERINFOA));
  return lpElement;
}

DELIMITERINFOW* StackInsertDelimiterW(HDELIMSTACK *hStack, int nDelimiterLen)
{
  DELIMITERINFOW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(DELIMITERINFOW));
  return lpElement;
}

DELIMITERINFOA* StackGetDelimiterA(HDELIMSTACK *hStack, char chDelimiter)
{
  DELIMITERINFOA *lpElement=(DELIMITERINFOA *)hStack->first;

  while (lpElement)
  {
    if (lpElement->pDelimiter[0] == chDelimiter)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

DELIMITERINFOW* StackGetDelimiterW(HDELIMSTACK *hStack, wchar_t wchDelimiter)
{
  DELIMITERINFOW *lpElement=(DELIMITERINFOW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->wpDelimiter[0] == wchDelimiter)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeDelimiterA(HDELIMSTACK *hStack)
{
  DELIMITERINFOA *lpElement=(DELIMITERINFOA *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->pDelimiter);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackFreeDelimiterW(HDELIMSTACK *hStack)
{
  DELIMITERINFOW *lpElement=(DELIMITERINFOW *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpDelimiter);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

WORDINFOA* StackInsertWordA(HWORDSTACK *hStack, int nWordLen)
{
  WORDINFOA *lpElement1;
  WORDINFOA *lpElement2=NULL;

  if ((DWORD)nWordLen < sizeof(hStack->lpWordLens) / sizeof(int))
  {
    if (hStack->lpWordLens[nWordLen])
    {
      lpElement1=(WORDINFOA *)hStack->lpWordLens[nWordLen];
    }
    else
    {
      lpElement1=(WORDINFOA *)hStack->first;

      while (lpElement1)
      {
        if (lpElement1->nWordLen >= nWordLen)
          break;

        lpElement1=lpElement1->next;
      }
    }
    StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(WORDINFOA));

    if (lpElement2)
      hStack->lpWordLens[nWordLen]=(int)lpElement2;
  }
  return lpElement2;
}

WORDINFOW* StackInsertWordW(HWORDSTACK *hStack, int nWordLen)
{
  WORDINFOW *lpElement1;
  WORDINFOW *lpElement2=NULL;

  if ((DWORD)nWordLen < sizeof(hStack->lpWordLens) / sizeof(int))
  {
    if (hStack->lpWordLens[nWordLen])
    {
      lpElement1=(WORDINFOW *)hStack->lpWordLens[nWordLen];
    }
    else
    {
      lpElement1=(WORDINFOW *)hStack->first;

      while (lpElement1)
      {
        if (lpElement1->nWordLen >= nWordLen)
          break;

        lpElement1=lpElement1->next;
      }
    }
    StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(WORDINFOW));

    if (lpElement2)
      hStack->lpWordLens[nWordLen]=(int)lpElement2;
  }
  return lpElement2;
}

WORDINFOA* StackGetWordA(HWORDSTACK *hStack, char *pWord, int nWordLen)
{
  WORDINFOA *lpElement=NULL;

  if ((DWORD)nWordLen < sizeof(hStack->lpWordLens) / sizeof(int))
  {
    lpElement=(WORDINFOA *)hStack->lpWordLens[nWordLen];

    while (lpElement)
    {
      if (lpElement->nWordLen != nWordLen)
        break;
      if (lpElement->dwFlags & AEHLF_MATCHCASE)
      {
        if (!lstrcmpA(lpElement->pWord, pWord))
          return lpElement;
      }
      else
      {
        if (!lstrcmpiA(lpElement->pWord, pWord))
          return lpElement;
      }
      lpElement=lpElement->next;
    }
  }
  return NULL;
}

WORDINFOW* StackGetWordW(HWORDSTACK *hStack, wchar_t *wpWord, int nWordLen)
{
  WORDINFOW *lpElement=NULL;

  if ((DWORD)nWordLen < sizeof(hStack->lpWordLens) / sizeof(int))
  {
    lpElement=(WORDINFOW *)hStack->lpWordLens[nWordLen];

    while (lpElement)
    {
      if (lpElement->nWordLen != nWordLen)
        break;
      if (lpElement->dwFlags & AEHLF_MATCHCASE)
      {
        if (!lstrcmpW(lpElement->wpWord, wpWord))
          return lpElement;
      }
      else
      {
        if (!lstrcmpiW(lpElement->wpWord, wpWord))
          return lpElement;
      }
      lpElement=lpElement->next;
    }
  }
  return NULL;
}

void StackFreeWordA(HWORDSTACK *hStack)
{
  WORDINFOA *lpElement=(WORDINFOA *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->pWord);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
  memset(hStack->lpWordLens, 0, MAX_PATH * sizeof(int));
}

void StackFreeWordW(HWORDSTACK *hStack)
{
  WORDINFOW *lpElement=(WORDINFOW *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpWord);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
  memset(hStack->lpWordLens, 0, MAX_PATH * sizeof(int));
}

QUOTEINFOA* StackInsertQuoteA(HQUOTESTACK *hStack, int nQuoteStartLen)
{
  QUOTEINFOA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(QUOTEINFOA));
  return lpElement;
}

QUOTEINFOW* StackInsertQuoteW(HQUOTESTACK *hStack, int nQuoteStartLen)
{
  QUOTEINFOW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(QUOTEINFOW));
  return lpElement;
}

void StackFreeQuoteA(HQUOTESTACK *hStack)
{
  QUOTEINFOA *lpElement=(QUOTEINFOA *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->pQuoteStart);
    GlobalFree((HGLOBAL)lpElement->pQuoteEnd);
    GlobalFree((HGLOBAL)lpElement->pQuoteInclude);
    GlobalFree((HGLOBAL)lpElement->pQuoteExclude);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackFreeQuoteW(HQUOTESTACK *hStack)
{
  QUOTEINFOW *lpElement=(QUOTEINFOW *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpQuoteStart);
    GlobalFree((HGLOBAL)lpElement->wpQuoteEnd);
    GlobalFree((HGLOBAL)lpElement->wpQuoteInclude);
    GlobalFree((HGLOBAL)lpElement->wpQuoteExclude);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

HWINDOWEDITA* StackInsertWindowA(HSTACK *hStack)
{
  HWINDOWEDITA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HWINDOWEDITA));

  return lpElement;
}

HWINDOWEDITW* StackInsertWindowW(HSTACK *hStack)
{
  HWINDOWEDITW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HWINDOWEDITW));

  return lpElement;
}

HWINDOWEDITA* StackGetWindowA(HSTACK *hStack, HWND hWnd)
{
  HWINDOWEDITA *lpElement=(HWINDOWEDITA *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HWINDOWEDITW* StackGetWindowW(HSTACK *hStack, HWND hWnd)
{
  HWINDOWEDITW *lpElement=(HWINDOWEDITW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackDeleteWindowA(HSTACK *hStack, HWINDOWEDITA *lpElement)
{
  RestoreFontAndColorsA(lpElement);
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement);
}

void StackDeleteWindowW(HSTACK *hStack, HWINDOWEDITW *lpElement)
{
  RestoreFontAndColorsW(lpElement);
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement);
}

void StackFreeWindowA(HSTACK *hStack)
{
  HWINDOWEDITA *lpElement=(HWINDOWEDITA *)hStack->first;
  HWINDOWEDITA *lpNextElement;

  while (lpElement)
  {
    lpNextElement=lpElement->next;
    StackDeleteWindowA(hStack, lpElement);
    lpElement=lpNextElement;
  }
}

void StackFreeWindowW(HSTACK *hStack)
{
  HWINDOWEDITW *lpElement=(HWINDOWEDITW *)hStack->first;
  HWINDOWEDITW *lpNextElement;

  while (lpElement)
  {
    lpNextElement=lpElement->next;
    StackDeleteWindowW(hStack, lpElement);
    lpElement=lpNextElement;
  }
}

MANUALSETA* StackInsertManualA(HSTACK *hStack)
{
  MANUALSETA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(MANUALSETA));

  return lpElement;
}

MANUALSETW* StackInsertManualW(HSTACK *hStack)
{
  MANUALSETW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(MANUALSETW));

  return lpElement;
}

MANUALSETA* StackGetManualA(HSTACK *hStack, HWND hWnd)
{
  MANUALSETA *lpElement=(MANUALSETA *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

MANUALSETW* StackGetManualW(HSTACK *hStack, HWND hWnd)
{
  MANUALSETW *lpElement=(MANUALSETW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackDeleteManual(HSTACK *hStack, void *ms)
{
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)ms);
}

MARKTHEME* StackInsertMarkTheme(HSTACK *hMarkThemesStack)
{
  MARKTHEME *lpMarkTheme=NULL;

  StackInsertIndex((stack **)&hMarkThemesStack->first, (stack **)&hMarkThemesStack->last, (stack **)&lpMarkTheme, -1, sizeof(MARKTHEME));
  return lpMarkTheme;
}

MARKTHEME* StackGetMarkTheme(HSTACK *hMarkThemesStack, HANDLE hTheme)
{
  MARKTHEME *lpMarkTheme=(MARKTHEME *)hMarkThemesStack->first;

  while (lpMarkTheme)
  {
    if (lpMarkTheme->hMarkThemeHandle == hTheme)
      return lpMarkTheme;

    lpMarkTheme=lpMarkTheme->next;
  }
  return NULL;
}

void StackDeleteMarkTheme(HSTACK *hMarkThemesStack, MARKTHEME *lpMarkTheme)
{
  StackDelete((stack **)&hMarkThemesStack->first, (stack **)&hMarkThemesStack->last, (stack *)lpMarkTheme);
}

void StackFreeMarkTheme(HWND hWnd, HSTACK *hMarkThemesStack)
{
  MARKTHEME *lpMarkTheme=(MARKTHEME *)hMarkThemesStack->first;

  while (lpMarkTheme)
  {
    StackFreeMarkText(hWnd, lpMarkTheme);

    lpMarkTheme=lpMarkTheme->next;
  }
  StackClear((stack **)&hMarkThemesStack->first, (stack **)&hMarkThemesStack->last);
}

MARKTEXT* StackInsertMarkText(HSTACK *hMarkTextsStack)
{
  MARKTEXT *lpMarkText=NULL;

  StackInsertIndex((stack **)&hMarkTextsStack->first, (stack **)&hMarkTextsStack->last, (stack **)&lpMarkText, -1, sizeof(MARKTEXT));
  return lpMarkText;
}

MARKTEXT* StackGetMarkText(HSTACK *hMarkTextsStack, DWORD dwMarkID)
{
  MARKTEXT *lpMarkText=(MARKTEXT *)hMarkTextsStack->first;

  while (lpMarkText)
  {
    if (lpMarkText->dwMarkID == dwMarkID)
      return lpMarkText;

    lpMarkText=lpMarkText->next;
  }
  return NULL;
}

BOOL StackIsAnyMarkText(HSTACK *hMarkTextsStack)
{
  if ((MARKTEXT *)hMarkTextsStack->first)
    return TRUE;
  return FALSE;
}

void StackDeleteMarkText(HWND hWnd, MARKTHEME *lpMarkTheme, MARKTEXT *lpMarkText)
{
  if (hWnd) SendMessage(hWnd, AEM_HLDELETEMARKTEXT, (WPARAM)lpMarkTheme->hMarkThemeHandle, (LPARAM)lpMarkText->hMarkTextHandle);
  StackDelete((stack **)&lpMarkTheme->hMarkTextsStack.first, (stack **)&lpMarkTheme->hMarkTextsStack.last, (stack *)lpMarkText);
}

void StackFreeMarkText(HWND hWnd, MARKTHEME *lpMarkTheme)
{
  MARKTEXT *lpMarkText=(MARKTEXT *)lpMarkTheme->hMarkTextsStack.first;
  MARKTEXT *lpNextMarkText;

  while (lpMarkText)
  {
    lpNextMarkText=lpMarkText->next;

    if (lpMarkText->hMarkTextHandle)
      StackDeleteMarkText(hWnd, lpMarkTheme, lpMarkText);

    lpMarkText=lpNextMarkText;
  }
}

BOOL MarkSelection(HWND hWnd, DWORD dwColorText, DWORD dwColorBk, BOOL bMatchCase, DWORD dwFontStyle, DWORD dwMarkID)
{
  CHARRANGE cr;
  wchar_t *wpText;
  int nTextLen=0;
  BOOL bResult=FALSE;

  if (dwMarkID)
  {
    SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);

    if (cr.cpMin != cr.cpMax && cr.cpMax - cr.cpMin < MAX_PATH)
    {
      if (wpText=(wchar_t *)SendMessage(hMainWnd, AKD_GETSELTEXTW, (WPARAM)hWnd, (LPARAM)&nTextLen))
      {
        MARKTHEME *lpMarkTheme=NULL;
        MARKTEXT *lpMarkText=NULL;
        HANDLE hMarkThemeHandle=NULL;
        HANDLE hMarkTextHandle=NULL;
        BOOL bCreated=FALSE;

        if (!(hMarkThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEW, 0, (LPARAM)NULL)))
        {
          hMarkThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLCREATETHEMEW, 0, (LPARAM)L"HighLight::MarkText");
          SendMessage(hWnd, AEM_HLSETTHEME, (WPARAM)hMarkThemeHandle, FALSE);
          bCreated=TRUE;
        }
        else
        {
          if (lpMarkTheme=StackGetMarkTheme(&hMarkThemeStack, hMarkThemeHandle))
          {
            if (lpMarkText=StackGetMarkText(&lpMarkTheme->hMarkTextsStack, dwMarkID))
            {
              if (lpMarkText->hMarkTextHandle)
                StackDeleteMarkText(hWnd, lpMarkTheme, lpMarkText);
            }
          }
        }

        if (hMarkThemeHandle)
        {
          if (!lpMarkTheme)
          {
            if (lpMarkTheme=StackInsertMarkTheme(&hMarkThemeStack))
            {
              lpMarkTheme->hMarkThemeHandle=hMarkThemeHandle;
              lpMarkTheme->bCreated=bCreated;
            }
          }

          if (lpMarkTheme)
          {
            AEMARKTEXTITEMW mti;

            mti.pMarkText=wpText;
            mti.nMarkTextLen=nTextLen;
            mti.dwFlags=bMatchCase?AEHLF_MATCHCASE:0;
            mti.crText=dwColorText;
            mti.crBk=dwColorBk;
            mti.dwFontStyle=dwFontStyle;
            if (hMarkTextHandle=(HANDLE)SendMessage(hWnd, AEM_HLADDMARKTEXTW, (WPARAM)lpMarkTheme->hMarkThemeHandle, (LPARAM)&mti))
              bResult=TRUE;

            if (lpMarkText=StackInsertMarkText(&lpMarkTheme->hMarkTextsStack))
            {
              lpMarkText->hMarkTextHandle=hMarkTextHandle;
              lpMarkText->dwMarkID=dwMarkID;
            }
          }
        }
        SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)wpText);
      }
    }
  }
  return bResult;
}

BOOL UnmarkSelection(HWND hWnd, DWORD dwMarkID)
{
  MARKTHEME *lpMarkTheme=NULL;
  MARKTEXT *lpMarkText=NULL;
  HANDLE hMarkThemeHandle=NULL;
  BOOL bResult=FALSE;

  if (hMarkThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEW, 0, (LPARAM)NULL))
  {
    if (lpMarkTheme=StackGetMarkTheme(&hMarkThemeStack, hMarkThemeHandle))
    {
      if (dwMarkID)
      {
        if (lpMarkText=StackGetMarkText(&lpMarkTheme->hMarkTextsStack, dwMarkID))
        {
          if (lpMarkText->hMarkTextHandle)
            StackDeleteMarkText(hWnd, lpMarkTheme, lpMarkText);
        }
      }
      else StackFreeMarkText(hWnd, lpMarkTheme);

      if (!StackIsAnyMarkText(&lpMarkTheme->hMarkTextsStack))
      {
        if (lpMarkTheme->bCreated)
          SendMessage(hWnd, AEM_HLDELETETHEME, (WPARAM)lpMarkTheme->hMarkThemeHandle, 0);
        StackDeleteMarkTheme(&hMarkThemeStack, lpMarkTheme);
      }
      bResult=TRUE;
    }
  }
  return bResult;
}

void CreateEditThemeA(HSYNTAXFILEA *lpSyntaxFileA, HWND hWnd)
{
  AEDELIMITEMA di;
  AEWORDITEMA wi;
  AEQUOTEITEMA qi;

  wsprintfA(szBuffer, "HighLight::File::%s", lpSyntaxFileA->szSyntaxFileName);

  if (lpSyntaxFileA->hThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLCREATETHEMEA, 0, (LPARAM)szBuffer))
  {
    //Transfer delimiters
    {
      DELIMITERINFOA *lpElement=(DELIMITERINFOA *)lpSyntaxFileA->hDelimiterStack.first;

      while (lpElement)
      {
        di.pDelimiter=lpElement->pDelimiter;
        di.nDelimiterLen=lpElement->nDelimiterLen;
        di.dwFlags=lpElement->dwFlags;
        di.crText=lpElement->dwColor1;
        di.crBk=lpElement->dwColor2;
        di.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;
        SendMessage(hWnd, AEM_HLADDDELIMITERA, (WPARAM)lpSyntaxFileA->hThemeHandle, (LPARAM)&di);

        lpElement=lpElement->next;
      }
    }

    //Transfer words
    {
      WORDINFOA *lpElement=(WORDINFOA *)lpSyntaxFileA->hWordStack.last;

      while (lpElement)
      {
        wi.pWord=lpElement->pWord;
        wi.nWordLen=lpElement->nWordLen;
        wi.dwFlags=lpElement->dwFlags;
        wi.crText=lpElement->dwColor1;
        wi.crBk=lpElement->dwColor2;
        wi.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;
        SendMessage(hWnd, AEM_HLADDWORDA, (WPARAM)lpSyntaxFileA->hThemeHandle, (LPARAM)&wi);

        lpElement=lpElement->prev;
      }
    }

    //Transfer quotes
    {
      QUOTEINFOA *lpElement=(QUOTEINFOA *)lpSyntaxFileA->hQuoteStack.first;

      while (lpElement)
      {
        qi.pQuoteStart=lpElement->pQuoteStart;
        qi.nQuoteStartLen=lpElement->nQuoteStartLen;
        qi.pQuoteEnd=lpElement->pQuoteEnd;
        qi.nQuoteEndLen=lpElement->nQuoteEndLen;
        qi.chEscape=lpElement->chEscape;
        qi.pQuoteInclude=lpElement->pQuoteInclude;
        qi.nQuoteIncludeLen=lpElement->nQuoteIncludeLen;
        qi.pQuoteExclude=lpElement->pQuoteExclude;
        qi.nQuoteExcludeLen=lpElement->nQuoteExcludeLen;
        qi.dwFlags=lpElement->dwFlags;
        qi.crText=lpElement->dwColor1;
        qi.crBk=lpElement->dwColor2;
        qi.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;

        if (qi.nQuoteIncludeLen)
          qi.dwFlags|=AEHLF_QUOTEINCLUDE;
        if (qi.nQuoteExcludeLen)
          qi.dwFlags|=AEHLF_QUOTEEXCLUDE;
        SendMessage(hWnd, AEM_HLADDQUOTEA, (WPARAM)lpSyntaxFileA->hThemeHandle, (LPARAM)&qi);

        lpElement=lpElement->next;
      }
    }
  }
}

void CreateEditThemeW(HSYNTAXFILEW *lpSyntaxFileW, HWND hWnd)
{
  AEDELIMITEMW di;
  AEWORDITEMW wi;
  AEQUOTEITEMW qi;

  wsprintfW(wszBuffer, L"HighLight::File::%s", lpSyntaxFileW->wszSyntaxFileName);

  if (lpSyntaxFileW->hThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLCREATETHEMEW, 0, (LPARAM)wszBuffer))
  {
    //Transfer delimiters
    {
      DELIMITERINFOW *lpElement=(DELIMITERINFOW *)lpSyntaxFileW->hDelimiterStack.first;

      while (lpElement)
      {
        di.pDelimiter=lpElement->wpDelimiter;
        di.nDelimiterLen=lpElement->nDelimiterLen;
        di.dwFlags=lpElement->dwFlags;
        di.crText=lpElement->dwColor1;
        di.crBk=lpElement->dwColor2;
        di.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;
        SendMessage(hWnd, AEM_HLADDDELIMITERW, (WPARAM)lpSyntaxFileW->hThemeHandle, (LPARAM)&di);

        lpElement=lpElement->next;
      }
    }

    //Transfer words
    {
      WORDINFOW *lpElement=(WORDINFOW *)lpSyntaxFileW->hWordStack.last;

      while (lpElement)
      {
        wi.pWord=lpElement->wpWord;
        wi.nWordLen=lpElement->nWordLen;
        wi.dwFlags=lpElement->dwFlags;
        wi.crText=lpElement->dwColor1;
        wi.crBk=lpElement->dwColor2;
        wi.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;
        SendMessage(hWnd, AEM_HLADDWORDW, (WPARAM)lpSyntaxFileW->hThemeHandle, (LPARAM)&wi);

        lpElement=lpElement->prev;
      }
    }

    //Transfer quotes
    {
      QUOTEINFOW *lpElement=(QUOTEINFOW *)lpSyntaxFileW->hQuoteStack.first;

      while (lpElement)
      {
        qi.pQuoteStart=lpElement->wpQuoteStart;
        qi.nQuoteStartLen=lpElement->nQuoteStartLen;
        qi.pQuoteEnd=lpElement->wpQuoteEnd;
        qi.nQuoteEndLen=lpElement->nQuoteEndLen;
        qi.chEscape=lpElement->wchEscape;
        qi.pQuoteInclude=lpElement->wpQuoteInclude;
        qi.nQuoteIncludeLen=lpElement->nQuoteIncludeLen;
        qi.pQuoteExclude=lpElement->wpQuoteExclude;
        qi.nQuoteExcludeLen=lpElement->nQuoteExcludeLen;
        qi.dwFlags=lpElement->dwFlags;
        qi.crText=lpElement->dwColor1;
        qi.crBk=lpElement->dwColor2;
        qi.dwFontStyle=bGlobalFontStyleEnable?nGlobalFontStyle:lpElement->dwFontStyle;

        if (qi.nQuoteIncludeLen)
          qi.dwFlags|=AEHLF_QUOTEINCLUDE;
        if (qi.nQuoteExcludeLen)
          qi.dwFlags|=AEHLF_QUOTEEXCLUDE;
        SendMessage(hWnd, AEM_HLADDQUOTEW, (WPARAM)lpSyntaxFileW->hThemeHandle, (LPARAM)&qi);

        lpElement=lpElement->next;
      }
    }
  }
}

void GetFontAndColorsA(HSYNTAXFILEA *lpSyntaxFileA, HWINDOWEDITA *lpWindow)
{
  if (!lpWindow->bOldColors)
  {
    lpWindow->bOldColors=TRUE;
    lpWindow->aecNew=lpSyntaxFileA->aecColors;

    if (lpWindow->aecNew.dwFlags)
    {
      lpWindow->aecOld.dwFlags=AECLR_ALL;
      SendMessage(lpWindow->hWnd, AEM_GETCOLORS, 0, (LPARAM)&lpWindow->aecOld);
    }
  }
  if (!lpWindow->bOldFont)
  {
    lpWindow->bOldFont=TRUE;
    lpWindow->dwFontFlags=lpSyntaxFileA->dwFontFlags;

    if (lpWindow->dwFontFlags)
    {
      LOGFONTA lfFontOld;

      lpWindow->hFontOld=(HFONT)SendMessage(lpWindow->hWnd, WM_GETFONT, 0, 0);
      GetObjectA(lpWindow->hFontOld, sizeof(LOGFONTA), &lfFontOld);

      if (!(lpWindow->dwFontFlags & FF_FONTSTYLE))
      {
        lpSyntaxFileA->lfFont.lfWeight=lfFontOld.lfWeight;
        lpSyntaxFileA->lfFont.lfItalic=lfFontOld.lfItalic;
      }
      if (!(lpWindow->dwFontFlags & FF_FONTSIZE))
      {
        lpSyntaxFileA->lfFont.lfHeight=-mod(lfFontOld.lfHeight);
      }
      if (!(lpWindow->dwFontFlags & FF_FONTFACE))
      {
        lstrcpynA(lpSyntaxFileA->lfFont.lfFaceName, lfFontOld.lfFaceName, LF_FACESIZE);
      }
      lpWindow->hFontNew=(HFONT)CreateFontIndirectA(&lpSyntaxFileA->lfFont);
    }
  }
}

void GetFontAndColorsW(HSYNTAXFILEW *lpSyntaxFileW, HWINDOWEDITW *lpWindow)
{
  if (!lpWindow->bOldColors)
  {
    lpWindow->bOldColors=TRUE;
    lpWindow->aecNew=lpSyntaxFileW->aecColors;

    if (lpWindow->aecNew.dwFlags)
    {
      lpWindow->aecOld.dwFlags=AECLR_ALL;
      SendMessage(lpWindow->hWnd, AEM_GETCOLORS, 0, (LPARAM)&lpWindow->aecOld);
    }
  }
  if (!lpWindow->bOldFont)
  {
    lpWindow->bOldFont=TRUE;
    lpWindow->dwFontFlags=lpSyntaxFileW->dwFontFlags;

    if (lpWindow->dwFontFlags)
    {
      LOGFONTW lfFontOld;

      lpWindow->hFontOld=(HFONT)SendMessage(lpWindow->hWnd, WM_GETFONT, 0, 0);
      GetObjectW(lpWindow->hFontOld, sizeof(LOGFONTW), &lfFontOld);

      if (!(lpWindow->dwFontFlags & FF_FONTSTYLE))
      {
        lpSyntaxFileW->lfFont.lfWeight=lfFontOld.lfWeight;
        lpSyntaxFileW->lfFont.lfItalic=lfFontOld.lfItalic;
      }
      if (!(lpWindow->dwFontFlags & FF_FONTSIZE))
      {
        lpSyntaxFileW->lfFont.lfHeight=-mod(lfFontOld.lfHeight);
      }
      if (!(lpWindow->dwFontFlags & FF_FONTFACE))
      {
        lstrcpynW(lpSyntaxFileW->lfFont.lfFaceName, lfFontOld.lfFaceName, LF_FACESIZE);
      }
      lpWindow->hFontNew=(HFONT)CreateFontIndirectW(&lpSyntaxFileW->lfFont);
    }
  }
}

void RestoreFontAndColorsA(HWINDOWEDITA *lpWindow)
{
  if (lpWindow->bOldColors)
  {
    lpWindow->bOldColors=FALSE;

    if (lpWindow->aecNew.dwFlags)
    {
      lpWindow->aecOld.dwFlags=GetColorsToRestore(lpWindow->hWnd, &lpWindow->aecNew);
      SendMessage(lpWindow->hWnd, AEM_SETCOLORS, 0, (LPARAM)&lpWindow->aecOld);
    }
  }
  if (lpWindow->bOldFont)
  {
    lpWindow->bOldFont=FALSE;

    if (lpWindow->dwFontFlags)
    {
      if (lpWindow->hFontNew == (HFONT)SendMessage(lpWindow->hWnd, WM_GETFONT, 0, 0))
        SendMessage(lpWindow->hWnd, WM_SETFONT, (WPARAM)lpWindow->hFontOld, TRUE);
      if (lpWindow->hFontNew)
      {
        DeleteObject(lpWindow->hFontNew);
        lpWindow->hFontNew=NULL;
      }
    }
  }
}

void RestoreFontAndColorsW(HWINDOWEDITW *lpWindow)
{
  if (lpWindow->bOldColors)
  {
    lpWindow->bOldColors=FALSE;

    if (lpWindow->aecNew.dwFlags)
    {
      lpWindow->aecOld.dwFlags=GetColorsToRestore(lpWindow->hWnd, &lpWindow->aecNew);
      SendMessage(lpWindow->hWnd, AEM_SETCOLORS, 0, (LPARAM)&lpWindow->aecOld);
    }
  }
  if (lpWindow->bOldFont)
  {
    lpWindow->bOldFont=FALSE;

    if (lpWindow->dwFontFlags)
    {
      if (lpWindow->hFontNew == (HFONT)SendMessage(lpWindow->hWnd, WM_GETFONT, 0, 0))
        SendMessage(lpWindow->hWnd, WM_SETFONT, (WPARAM)lpWindow->hFontOld, TRUE);
      if (lpWindow->hFontNew)
      {
        DeleteObject(lpWindow->hFontNew);
        lpWindow->hFontNew=NULL;
      }
    }
  }
}

void UnassignThemeA(HWND hWnd)
{
  HANDLE hThemeHandle;
  wchar_t wszThemeName[MAX_PATH];

  if (hThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEA, 0, (LPARAM)NULL))
  {
    if (SendMessage(hWnd, AEM_HLGETTHEMENAMEW, (WPARAM)hThemeHandle, (LPARAM)wszThemeName))
    {
      if (!xstrcmpnW(L"HighLight::File::", wszThemeName, (DWORD)-1))
      {
        UnmarkSelection(hWnd, 0);
        SendMessage(hWnd, AEM_HLSETTHEME, (WPARAM)NULL, FALSE);
      }
    }
  }
}

void UnassignThemeW(HWND hWnd)
{
  HANDLE hThemeHandle;
  wchar_t wszThemeName[MAX_PATH];

  if (hThemeHandle=(HANDLE)SendMessage(hWnd, AEM_HLGETTHEMEW, 0, (LPARAM)NULL))
  {
    if (SendMessage(hWnd, AEM_HLGETTHEMENAMEW, (WPARAM)hThemeHandle, (LPARAM)wszThemeName))
    {
      if (!xstrcmpnW(L"HighLight::File::", wszThemeName, (DWORD)-1))
      {
        UnmarkSelection(hWnd, 0);
        SendMessage(hWnd, AEM_HLSETTHEME, (WPARAM)NULL, FALSE);
      }
    }
  }
}

void ClearCache()
{
  if (bInitMain)
  {
    BOOL bTmp;

    bTmp=bUseCache;
    UninitMain();
    bUseCache=FALSE;
    InitMain();
    bUseCache=bTmp;

    bUpdateTheme=TRUE;
    UpdateEditAll();
  }
}

void UpdateEdit(HWND hWnd)
{
  RECT rcDraw;

  SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcDraw);
  InvalidateRect(hWnd, &rcDraw, FALSE);
}

void UpdateEditAll()
{
  HWND hWndFrame;
  HWND hWndChild;

  if (!bMDI)
  {
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
        UpdateEdit(hWndChild);
      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }
  }
  else
  {
    hWndFrame=GetWindow(hMdiClient, GW_CHILD);

    while (hWndFrame)
    {
      hWndChild=GetWindow(hWndFrame, GW_CHILD);

      while (hWndChild)
      {
        if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
          UpdateEdit(hWndChild);
        hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
      }
      hWndFrame=GetNextWindow(hWndFrame, GW_HWNDNEXT);
    }
  }
}

void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt, TEXTMETRICA *tm)
{
  DWORD dwPosition;

  if (bOldRichEdit)
  {
    dwPosition=SendMessage(hWnd, EM_POSFROMCHAR, nCharIndex, (LPARAM)NULL);
    pt->x=LOWORD(dwPosition);
    pt->y=HIWORD(dwPosition);

    if (tm)
    {
      if (pt->x > 0x10000 - tm->tmMaxCharWidth)
        pt->x-=0x10000;
      if (pt->y > 0x10000 - tm->tmHeight)
        pt->y-=0x10000;
    }
  }
  else
  {
    SendMessage(hWnd, EM_POSFROMCHAR, (WPARAM)pt, nCharIndex);
  }
}

int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote)
{
  char *pCount;
  char chStopChar;
  int nWordLen;

  while (*pText == ' ' || *pText == '\t') ++pText;

  if (*pText == '\"' || *pText == '\'' || *pText == '`')
  {
    //Parse: "param" or 'param' or `param`
    chStopChar=*pText;
    pCount=++pText;
    while (*pCount != chStopChar && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
      if (*pCount == chStopChar)
        ++*pNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    pCount=pText;

    while (*pCount != ' ' && *pCount != '\t' && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(pCount - pText, nWordLenMax - 1);
  memcpy(szWord, pText, nWordLen);
  szWord[nWordLen]='\0';

  return nWordLen;
}

int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote)
{
  wchar_t *wpCount;
  wchar_t wchStopChar;
  int nWordLen;

  while (*wpText == ' ' || *wpText == '\t') ++wpText;

  if (*wpText == '\"' || *wpText == '\'' || *wpText == '`')
  {
    //Parse: "param" or 'param' or `param`
    wchStopChar=*wpText;
    wpCount=++wpText;
    while (*wpCount != wchStopChar && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
      if (*wpCount == wchStopChar)
        ++*wpNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    wpCount=wpText;

    while (*wpCount != ' ' && *wpCount != '\t' && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(wpCount - wpText, nWordLenMax - 1);
  memcpy(wszWord, wpText, nWordLen * sizeof(wchar_t));
  wszWord[nWordLen]='\0';

  return nWordLen;
}

BOOL NextLineA(char **pText)
{
  while (**pText != '\r' && **pText != '\0') ++*pText;
  if (**pText == '\0') return FALSE;
  if (*++*pText == '\n') ++*pText;
  return TRUE;
}

BOOL NextLineW(wchar_t **wpText)
{
  while (**wpText != '\r' && **wpText != '\0') ++*wpText;
  if (**wpText == '\0') return FALSE;
  if (*++*wpText == '\n') ++*wpText;
  return TRUE;
}

BOOL SkipCommentA(char **pText)
{
  while (1)
  {
    while (**pText == ' ' || **pText == '\t' || **pText == '\r' || **pText == '\n') ++*pText;

    if (**pText == ';')
    {
      while (**pText != '\r' && **pText != '\0') ++*pText;
    }
    else break;
  }
  if (**pText == '\0')
    return FALSE;
  return TRUE;
}

BOOL SkipCommentW(wchar_t **wpText)
{
  while (1)
  {
    while (**wpText == ' ' || **wpText == '\t' || **wpText == '\r' || **wpText == '\n') ++*wpText;

    if (**wpText == ';')
    {
      while (**wpText != '\r' && **wpText != '\0') ++*wpText;
    }
    else break;
  }
  if (**wpText == '\0')
    return FALSE;
  return TRUE;
}

int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenA(pFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (pFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (pFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynA(szBaseName, pFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenW(wpFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (wpFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (wpFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynW(wszBaseName, wpFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

char* GetFileExtA(char *pFile)
{
  int i;

  for (i=lstrlenA(pFile) - 1; i >= 0; --i)
  {
    if (pFile[i] == '.') return (pFile + i + 1);
    else if (pFile[i] == '\\') break;
  }
  return "";
}

wchar_t* GetFileExtW(wchar_t *wpFile)
{
  int i;

  for (i=lstrlenW(wpFile) - 1; i >= 0; --i)
  {
    if (wpFile[i] == '.') return (wpFile + i + 1);
    else if (wpFile[i] == '\\') break;
  }
  return L"";
}

DWORD GetColorsToRestore(HWND hWnd, AECOLORS *aecHighlight)
{
  AECOLORS aecCurrent;
  DWORD dwFlags=aecHighlight->dwFlags;

  aecCurrent.dwFlags=AECLR_ALL;
  SendMessage(hWnd, AEM_GETCOLORS, 0, (LPARAM)&aecCurrent);

  if (dwFlags & AECLR_CARET)
  {
    if (aecHighlight->crCaret != aecCurrent.crCaret)
      dwFlags&=~AECLR_CARET;
  }
  if (dwFlags & AECLR_BASICTEXT)
  {
    if (aecHighlight->crBasicText != aecCurrent.crBasicText)
      dwFlags&=~AECLR_BASICTEXT;
  }
  if (dwFlags & AECLR_BASICBK)
  {
    if (aecHighlight->crBasicBk != aecCurrent.crBasicBk)
      dwFlags&=~AECLR_BASICBK;
  }
  if (dwFlags & AECLR_SELTEXT)
  {
    if (aecHighlight->crSelText != aecCurrent.crSelText)
      dwFlags&=~AECLR_SELTEXT;
  }
  if (dwFlags & AECLR_SELBK)
  {
    if (aecHighlight->crSelBk != aecCurrent.crSelBk)
      dwFlags&=~AECLR_SELBK;
  }
  if (dwFlags & AECLR_ACTIVELINETEXT)
  {
    if (aecHighlight->crActiveLineText != aecCurrent.crActiveLineText)
      dwFlags&=~AECLR_ACTIVELINETEXT;
  }
  if (dwFlags & AECLR_ACTIVELINEBK)
  {
    if (aecHighlight->crActiveLineBk != aecCurrent.crActiveLineBk)
      dwFlags&=~AECLR_ACTIVELINEBK;
  }
  if (dwFlags & AECLR_URLTEXT)
  {
    if (aecHighlight->crUrlText != aecCurrent.crUrlText)
      dwFlags&=~AECLR_URLTEXT;
  }
  if (dwFlags & AECLR_ACTIVECOLUMN)
  {
    if (aecHighlight->crActiveColumn != aecCurrent.crActiveColumn)
      dwFlags&=~AECLR_ACTIVECOLUMN;
  }
  if (dwFlags & AECLR_COLUMNMARKER)
  {
    if (aecHighlight->crColumnMarker != aecCurrent.crColumnMarker)
      dwFlags&=~AECLR_COLUMNMARKER;
  }
  return dwFlags;
}

COLORREF GetColorValueFromStrA(char *pColor)
{
  COLORREF crColor;

  crColor=hex2decA(pColor);
  crColor=RGB(GetBValue(crColor), GetGValue(crColor), GetRValue(crColor));
  return crColor;
}

COLORREF GetColorValueFromStrW(wchar_t *wpColor)
{
  COLORREF crColor;

  crColor=hex2decW(wpColor);
  crColor=RGB(GetBValue(crColor), GetGValue(crColor), GetRValue(crColor));
  return crColor;
}

char* GetColorStrFromValueA(COLORREF crColor, char *szColor)
{
  wsprintfA(szColor, "%02X%02X%02X", GetRValue(crColor), GetGValue(crColor), GetBValue(crColor));
  return szColor;
}

wchar_t* GetColorStrFromValueW(COLORREF crColor, wchar_t *wszColor)
{
  wsprintfW(wszColor, L"%02X%02X%02X", GetRValue(crColor), GetGValue(crColor), GetBValue(crColor));
  return wszColor;
}


//// Options

void ReadSyntaxFilesA()
{
  WIN32_FIND_DATAA wfd;
  INIVALUEA iv;
  HANDLE hIniFile=NULL;
  HANDLE hSearch;
  char szFindFiles[MAX_PATH];
  char szHighLightFile[MAX_PATH];
  char szExtensions[MAX_PATH];
  BOOL bSyntaxFilesCache;

  //Open cache file
  wsprintfA(szBuffer, "%s\\cache", szHighLightDir);
  hIniFile=(HANDLE)SendMessage(hMainWnd, AKD_INIOPEN, POB_READ, (LPARAM)szBuffer);

  //Start search for *.highlight
  wsprintfA(szFindFiles, "%s\\*.highlight", szHighLightDir);

  if ((hSearch=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      bSyntaxFilesCache=FALSE;

      //Read cache
      if (bUseCache && hIniFile)
      {
        iv.pSection="Cache";
        iv.pKey=wfd.cFileName;
        iv.dwType=INI_STRINGANSI;
        iv.lpData=(LPBYTE)szExtensions;
        iv.dwData=MAX_PATH;

        if (SendMessage(hMainWnd, AKD_INIGETVALUE, (WPARAM)hIniFile, (LPARAM)&iv))
        {
          HSYNTAXFILEA *lpSyntaxFile=NULL;
          EXTENSIONINFOA *lpExtElement=NULL;
          char *pText=szExtensions;
          char *pExtension;
          int nExtensionLen;
          BOOL bQuoteString;

          if (lpSyntaxFile=StackInsertSyntaxFileA(&hSyntaxFilesStack))
          {
            GetBaseNameA(wfd.cFileName, lpSyntaxFile->szSyntaxFileName, MAX_PATH);
            lpSyntaxFile->bCache=TRUE;

            //Parse line
            while (1)
            {
              lpExtElement=NULL;
              pExtension=NULL;

              //Extension
              if ((nExtensionLen=GetWordA(pText, szBuffer, MAX_PATH, &pText, &bQuoteString)) || bQuoteString)
              {
                if (pExtension=(char *)GlobalAlloc(GPTR, nExtensionLen + 1))
                  memcpy(pExtension, szBuffer, nExtensionLen + 1);
                else
                  break;
              }
              else break;

              //Add to stack
              if (lpExtElement=StackInsertExtensionA(&lpSyntaxFile->hExtensionStack, nExtensionLen))
              {
                lpExtElement->pExtension=pExtension;
                lpExtElement->nExtensionLen=nExtensionLen;
              }
            }
          }
          bSyntaxFilesCache=TRUE;
        }
      }

      if (!bSyntaxFilesCache)
      {
        wsprintfA(szHighLightFile, "%s\\%s", szHighLightDir, wfd.cFileName);
        StackLoadSyntaxFileA(szHighLightFile, &hSyntaxFilesStack, NULL);
        bSaveOptions=TRUE;
      }
    }
    while (FindNextFileA(hSearch, &wfd));

    FindClose(hSearch);
  }
  if (hIniFile) SendMessage(hMainWnd, AKD_INICLOSE, (WPARAM)hIniFile, 0);
}

void ReadSyntaxFilesW()
{
  WIN32_FIND_DATAW wfd;
  INIVALUEW iv;
  HANDLE hIniFile=NULL;
  HANDLE hSearch;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszHighLightFile[MAX_PATH];
  wchar_t wszExtensions[MAX_PATH];
  BOOL bSyntaxFilesCache;

  //Open cache file
  wsprintfW(wszBuffer, L"%s\\cache", wszHighLightDir);
  hIniFile=(HANDLE)SendMessage(hMainWnd, AKD_INIOPEN, POB_READ, (LPARAM)wszBuffer);

  //Start search for *.highlight
  wsprintfW(wszFindFiles, L"%s\\*.highlight", wszHighLightDir);

  if ((hSearch=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      bSyntaxFilesCache=FALSE;

      //Read cache
      if (bUseCache && hIniFile)
      {
        iv.pSection=L"Cache";
        iv.pKey=wfd.cFileName;
        iv.dwType=INI_STRINGUNICODE;
        iv.lpData=(LPBYTE)wszExtensions;
        iv.dwData=MAX_PATH * sizeof(wchar_t);

        if (SendMessage(hMainWnd, AKD_INIGETVALUE, (WPARAM)hIniFile, (LPARAM)&iv))
        {
          HSYNTAXFILEW *lpSyntaxFile=NULL;
          EXTENSIONINFOW *lpExtElement=NULL;
          wchar_t *wpText=wszExtensions;
          wchar_t *wpExtension;
          int nExtensionLen;
          BOOL bQuoteString;

          if (lpSyntaxFile=StackInsertSyntaxFileW(&hSyntaxFilesStack))
          {
            GetBaseNameW(wfd.cFileName, lpSyntaxFile->wszSyntaxFileName, MAX_PATH);
            lpSyntaxFile->bCache=TRUE;

            //Parse line
            while (1)
            {
              lpExtElement=NULL;
              wpExtension=NULL;

              //Extension
              if ((nExtensionLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
              {
                if (wpExtension=(wchar_t *)GlobalAlloc(GPTR, nExtensionLen * sizeof(wchar_t) + 2))
                  memcpy(wpExtension, wszBuffer, nExtensionLen * sizeof(wchar_t) + 2);
                else
                  break;
              }
              else break;

              //Add to stack
              if (lpExtElement=StackInsertExtensionW(&lpSyntaxFile->hExtensionStack, nExtensionLen))
              {
                lpExtElement->wpExtension=wpExtension;
                lpExtElement->nExtensionLen=nExtensionLen;
              }
            }
          }
          bSyntaxFilesCache=TRUE;
        }
      }

      if (!bSyntaxFilesCache)
      {
        wsprintfW(wszHighLightFile, L"%s\\%s", wszHighLightDir, wfd.cFileName);
        StackLoadSyntaxFileW(wszHighLightFile, &hSyntaxFilesStack, NULL);
        bSaveOptions=TRUE;
      }
    }
    while (FindNextFileW(hSearch, &wfd));

    FindClose(hSearch);
  }
  if (hIniFile) SendMessage(hMainWnd, AKD_INICLOSE, (WPARAM)hIniFile, 0);
}

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  //Read settings
  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="GlobalFontStyleEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bGlobalFontStyleEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="GlobalFontStyle";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nGlobalFontStyle;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoMarkEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bAutoMarkEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoMarkType";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nAutoMarkType;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoMarkTextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAutoMarkTextColor;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoMarkBkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAutoMarkBkColor;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="UseCache";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bUseCache;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  //Read settings
  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"GlobalFontStyleEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bGlobalFontStyleEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"GlobalFontStyle";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nGlobalFontStyle;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoMarkEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bAutoMarkEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoMarkType";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nAutoMarkType;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoMarkTextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAutoMarkTextColor;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoMarkBkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAutoMarkBkColor;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"UseCache";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bUseCache;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (dwFlags & OF_SETTINGS)
  {
    if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
    {
      po.pOptionName="GlobalFontStyleEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bGlobalFontStyleEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="GlobalFontStyle";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nGlobalFontStyle;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoMarkEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bAutoMarkEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoMarkType";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nAutoMarkType;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoMarkTextColor";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAutoMarkTextColor;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoMarkBkColor";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAutoMarkBkColor;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="UseCache";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bUseCache;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
    }
  }

  if (dwFlags & OF_CACHE)
  {
    if (bUseCache)
    {
      HSYNTAXFILEA *lpSyntaxFile;
      EXTENSIONINFOA *lpExtElement;
      INIVALUEA iv;
      HANDLE hIniFile;
      HANDLE hIniSection;
      char szExtensions[MAX_PATH];
      int nExtensionsLen;

      //Open cache file
      wsprintfA(szBuffer, "%s\\cache", szHighLightDir);
      if (hIniFile=(HANDLE)SendMessage(hMainWnd, AKD_INIOPEN, POB_SAVE, (LPARAM)szBuffer))
      {
        //Clear cache section
        if (hIniSection=(HANDLE)SendMessage(hMainWnd, AKD_INIGETSECTION, (WPARAM)hIniFile, (LPARAM)"Cache"))
          SendMessage(hMainWnd, AKD_INICLEARSECTION, (WPARAM)hIniFile, (LPARAM)hIniSection);

        //Scan themes
        lpSyntaxFile=(HSYNTAXFILEA *)hSyntaxFilesStack.first;

        while (lpSyntaxFile)
        {
          szExtensions[0]='\0';
          nExtensionsLen=0;

          //Get extentions
          lpExtElement=(EXTENSIONINFOA *)lpSyntaxFile->hExtensionStack.first;

          while (lpExtElement)
          {
            if (nExtensionsLen + lstrlenA(lpExtElement->pExtension) >= MAX_PATH)
              break;
            nExtensionsLen=wsprintfA(szExtensions, "%s \"%s\"", szExtensions, lpExtElement->pExtension);
            lpExtElement=lpExtElement->next;
          }

          //Save extentions
          wsprintfA(szBuffer, "%s.highlight", lpSyntaxFile->szSyntaxFileName);
          iv.pSection="Cache";
          iv.pKey=szBuffer;
          iv.dwType=INI_STRINGANSI;
          iv.lpData=(LPBYTE)(szExtensions + 1);
          iv.dwData=nExtensionsLen;
          SendMessage(hMainWnd, AKD_INISETVALUE, (WPARAM)hIniFile, (LPARAM)&iv);

          lpSyntaxFile=lpSyntaxFile->next;
        }
        SendMessage(hMainWnd, AKD_INICLOSE, (WPARAM)hIniFile, 0);
      }
    }
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (dwFlags & OF_SETTINGS)
  {
    if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
    {
      po.pOptionName=L"GlobalFontStyleEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bGlobalFontStyleEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"GlobalFontStyle";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nGlobalFontStyle;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoMarkEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bAutoMarkEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoMarkType";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nAutoMarkType;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoMarkTextColor";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAutoMarkTextColor;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoMarkBkColor";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAutoMarkBkColor;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"UseCache";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bUseCache;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
    }
  }

  if (dwFlags & OF_CACHE)
  {
    if (bUseCache)
    {
      HSYNTAXFILEW *lpSyntaxFile;
      EXTENSIONINFOW *lpExtElement;
      INIVALUEW iv;
      HANDLE hIniFile;
      HANDLE hIniSection;
      wchar_t wszExtensions[MAX_PATH];
      int nExtensionsLen;

      //Open cache file
      wsprintfW(wszBuffer, L"%s\\cache", wszHighLightDir);
      if (hIniFile=(HANDLE)SendMessage(hMainWnd, AKD_INIOPEN, POB_SAVE, (LPARAM)wszBuffer))
      {
        //Clear cache
        if (hIniSection=(HANDLE)SendMessage(hMainWnd, AKD_INIGETSECTION, (WPARAM)hIniFile, (LPARAM)L"Cache"))
          SendMessage(hMainWnd, AKD_INICLEARSECTION, (WPARAM)hIniFile, (LPARAM)hIniSection);

        //Scan themes
        lpSyntaxFile=(HSYNTAXFILEW *)hSyntaxFilesStack.first;

        while (lpSyntaxFile)
        {
          wszExtensions[0]='\0';
          nExtensionsLen=0;

          //Get extentions
          lpExtElement=(EXTENSIONINFOW *)lpSyntaxFile->hExtensionStack.first;

          while (lpExtElement)
          {
            if (nExtensionsLen + lstrlenW(lpExtElement->wpExtension) >= MAX_PATH)
              break;
            nExtensionsLen=wsprintfW(wszExtensions, L"%s \"%s\"", wszExtensions, lpExtElement->wpExtension);
            lpExtElement=lpExtElement->next;
          }

          //Save extentions
          wsprintfW(wszBuffer, L"%s.highlight", lpSyntaxFile->wszSyntaxFileName);
          iv.pSection=L"Cache";
          iv.pKey=wszBuffer;
          iv.dwType=INI_STRINGUNICODE;
          iv.lpData=(LPBYTE)(wszExtensions + 1);
          iv.dwData=nExtensionsLen * sizeof(wchar_t);
          SendMessage(hMainWnd, AKD_INISETVALUE, (WPARAM)hIniFile, (LPARAM)&iv);

          lpSyntaxFile=lpSyntaxFile->next;
        }
        SendMessage(hMainWnd, AKD_INICLOSE, (WPARAM)hIniFile, 0);
      }
    }
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_GLOBAL_FONTSTYLE_GROUP)
      return L"\x0413\x043B\x043E\x0431\x0430\x043B\x044C\x043D\x044B\x0439\x0020\x0441\x0442\x0438\x043B\x044C\x0020\x0448\x0440\x0438\x0444\x0442\x0430";
    if (nStringID == STRID_ENABLE)
      return L"\x0412\x043A\x043B\x044E\x0447\x0435\x043D\x043E";
    if (nStringID == STRID_IGNORE)
      return L"\x0438\x0433\x043D\x043E\x0440\x0438\x0440\x043E\x0432\x0430\x0442\x044C";
    if (nStringID == STRID_NORMAL)
      return L"\x041E\x0431\x044B\x0447\x043D\x044B\x0439";
    if (nStringID == STRID_BOLD)
      return L"\x0416\x0438\x0440\x043D\x044B\x0439";
    if (nStringID == STRID_ITALIC)
      return L"\x041A\x0443\x0440\x0441\x0438\x0432";
    if (nStringID == STRID_BOLDITALIC)
      return L"\x0416\x0438\x0440\x043D\x044B\x0439\x0020\x041A\x0443\x0440\x0441\x0438\x0432";
    if (nStringID == STRID_AUTOMARK_GROUP)
      return L"\x0410\x0432\x0442\x043E\x043C\x0430\x0442\x0438\x0447\x0435\x0441\x043A\x0438\x0020\x043E\x0442\x043C\x0435\x0447\x0430\x0442\x044C\x0020\x0432\x044B\x0434\x0435\x043B\x0435\x043D\x043D\x043E\x0435";
    if (nStringID == STRID_SYMBOLS)
      return L"\x0421\x0438\x043C\x0432\x043E\x043B\x044B";
    if (nStringID == STRID_WORDS)
      return L"\x0421\x043B\x043E\x0432\x0430";
    if (nStringID == STRID_TEXTCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0442\x0435\x043A\x0441\x0442\x0430\x0020\x0028\x0052\x0052\x0047\x0047\x0042\x0042\x0029\x003A";
    if (nStringID == STRID_BGCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0444\x043E\x043D\x0430\x0020\x0028\x0052\x0052\x0047\x0047\x0042\x0042\x0029\x003A";
    if (nStringID == STRID_CACHE_ENABLE)
      return L"\x0418\x0441\x043F\x043E\x043B\x044C\x0437\x043E\x0432\x0430\x0442\x044C\x0020\x043A\x044D\x0448";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_GLOBAL_FONTSTYLE_GROUP)
      return L"Global font style";
    if (nStringID == STRID_ENABLE)
      return L"Enable";
    if (nStringID == STRID_IGNORE)
      return L"ignore";
    if (nStringID == STRID_NORMAL)
      return L"Normal";
    if (nStringID == STRID_BOLD)
      return L"Bold";
    if (nStringID == STRID_ITALIC)
      return L"Italic";
    if (nStringID == STRID_BOLDITALIC)
      return L"Bold Italic";
    if (nStringID == STRID_AUTOMARK_GROUP)
      return L"Automatic selection marking";
    if (nStringID == STRID_SYMBOLS)
      return L"Symbols";
    if (nStringID == STRID_WORDS)
      return L"Words";
    if (nStringID == STRID_TEXTCOLOR)
      return L"Text color (RRGGBB):";
    if (nStringID == STRID_BGCOLOR)
      return L"Background color (RRGGBB):";
    if (nStringID == STRID_CACHE_ENABLE)
      return L"Use cache";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMainIcon=pd->hMainIcon;
  hMdiClient=pd->hMdiClient;
  bOldWindows=pd->bOldWindows;
  bOldRichEdit=pd->bOldRichEdit;
  bAkelEdit=pd->bAkelEdit;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szHighLightDir, "%s\\AkelFiles\\Plugs\\HighLight", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszHighLightDir, L"%s\\AkelFiles\\Plugs\\HighLight", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  if (bOldWindows)
    ReadSyntaxFilesA();
  else
    ReadSyntaxFilesW();

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewFrameProcData);
  }

  NewEditProcData=NULL;
  SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)(bOldWindows?NewEditProcA:NewEditProcW), (LPARAM)&NewEditProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (bMDI)
  {
    if (NewFrameProcData)
    {
      SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
      NewFrameProcData=NULL;
    }
  }
  if (NewEditProcData)
  {
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
    NewEditProcData=NULL;
  }

  if (bOldWindows)
  {
    if (bSaveOptions) SaveOptionsA(OF_CACHE);
    StackFreeWindowA(&hWindowsStack);
    StackFreeMarkTheme(NULL, &hMarkThemeStack);
    StackFreeSyntaxFileA(&hSyntaxFilesStack);
  }
  else
  {
    if (bSaveOptions) SaveOptionsW(OF_CACHE);
    StackFreeWindowW(&hWindowsStack);
    StackFreeMarkTheme(NULL, &hMarkThemeStack);
    StackFreeSyntaxFileW(&hSyntaxFilesStack);
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
